# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the task to generate indexes for a suite."""

import datetime as dt
import hashlib
import lzma
from pathlib import PurePath
from textwrap import dedent
from typing import Any, ClassVar, override

from django_pglocks import advisory_lock

from debusine.artifacts.models import CollectionCategory
from debusine.db.locks import LockError, LockType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    File,
    WorkRequest,
)
from debusine.db.playground import scenarios
from debusine.server.tasks import GenerateSuiteIndexes
from debusine.server.tasks.generate_suite_indexes import IndexInfo
from debusine.server.tasks.models import (
    GenerateSuiteIndexesData,
    GenerateSuiteIndexesDynamicData,
)
from debusine.server.tasks.tests.helpers import ServerTaskTestCase
from debusine.tasks import TaskConfigError
from debusine.tasks.models import LookupSingle


class GenerateSuiteIndexesTestsBase(ServerTaskTestCase[GenerateSuiteIndexes]):
    """Tests for :py:class:`GenerateSuiteIndexes`."""

    def create_work_request(
        self,
        *,
        suite_collection: LookupSingle,
        generate_at: dt.datetime | None = None,
        assign_new_worker: bool = True,
    ) -> WorkRequest:
        """Create a :py:class:`GenerateSuiteIndexes` work request."""
        if generate_at is None:
            generate_at = dt.datetime.now(dt.UTC)
        work_request = self.playground.create_server_task(
            task_name="generatesuiteindexes",
            task_data=GenerateSuiteIndexesData(
                suite_collection=suite_collection, generate_at=generate_at
            ),
        )
        self.playground.advance_work_request(
            work_request, assign_new_worker=assign_new_worker
        )
        return work_request

    def create_generatesuiteindexes(
        self,
        *,
        suite_collection: LookupSingle,
        generate_at: dt.datetime | None = None,
        assign_new_worker: bool = True,
    ) -> GenerateSuiteIndexes:
        """Create an instance of the :py:class:`GenerateSuiteIndexes` task."""
        work_request = self.create_work_request(
            suite_collection=suite_collection,
            generate_at=generate_at,
            assign_new_worker=assign_new_worker,
        )
        task = work_request.get_task()
        assert isinstance(task, GenerateSuiteIndexes)
        return task


class GenerateSuiteIndexesTests(GenerateSuiteIndexesTestsBase):
    """Tests for :py:class:`GenerateSuiteIndexes`."""

    SAMPLE_TASK_DATA = {
        "suite_collection": f"bookworm@{CollectionCategory.SUITE}",
        "generate_at": dt.datetime.now(dt.UTC),
    }

    bookworm: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.bookworm = cls.create_suite_collection("bookworm")

    @classmethod
    def create_suite_collection(
        cls, name: str, data: dict[str, Any] | None = None
    ) -> Collection:
        """Create a `debian:suite` collection."""
        return cls.playground.create_collection(
            name=name, category=CollectionCategory.SUITE, data=data
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task, strip_task_inputs=True),
            GenerateSuiteIndexesDynamicData(
                suite_collection_id=self.bookworm.pk
            ),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """``get_input_artifacts_ids`` returns the empty list."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

    def test_suite(self) -> None:
        """``suite`` returns the requested suite collection."""
        self.assertEqual(self.task.suite.collection, self.bookworm)

    def test_components(self) -> None:
        """``components`` returns the configured components."""
        self.bookworm.data = {"components": ["main", "contrib"]}
        self.bookworm.save()
        self.assertEqual(self.task.components, ["main", "contrib"])

    def test_components_unset(self) -> None:
        """``components`` raises exception: components not configured."""
        with self.assertRaisesRegex(
            TaskConfigError,
            f"Collection '{self.task_data["suite_collection"]}'"
            " has no 'components' configuration",
        ):
            self.task.components

    def test_components_not_list(self) -> None:
        """``components`` raises exception: components not a list."""
        self.bookworm.data = {"components": "not-a-list"}
        self.bookworm.save()
        with self.assertRaisesRegex(
            TaskConfigError, "Expected 'components' as a list; got 'not-a-list'"
        ):
            self.task.components

    def test_architectures(self) -> None:
        """``architectures`` returns the configured architectures."""
        self.bookworm.data = {"architectures": ["amd64", "i386"]}
        self.bookworm.save()
        self.assertEqual(self.task.architectures, ["amd64", "i386"])

    def test_architectures_unset(self) -> None:
        """``architectures`` raises exception: architectures not configured."""
        with self.assertRaisesRegex(
            TaskConfigError,
            f"Collection '{self.task_data["suite_collection"]}'"
            " has no 'architectures' configuration",
        ):
            self.task.architectures

    def test_architectures_not_list(self) -> None:
        """``architectures`` raises exception: architectures not a list."""
        self.bookworm.data = {"architectures": "not-a-list"}
        self.bookworm.save()
        with self.assertRaisesRegex(
            TaskConfigError,
            "Expected 'architectures' as a list; got 'not-a-list'",
        ):
            self.task.architectures

    def test_supersede_older_items(self) -> None:
        """Remove superseded old items before generating new ones."""
        bookworm = self.bookworm
        suite_collection = f"bookworm@{CollectionCategory.SUITE}"
        trixie = self.create_suite_collection("trixie")
        dts = [
            dt.datetime(2025, month, 1, tzinfo=dt.UTC) for month in (1, 2, 3, 4)
        ]
        to_remove = [
            (bookworm, "index:main/source/Sources.xz"),
            (bookworm, "index:main/binary-amd64/Packages.xz"),
            (bookworm, "index:Release"),
        ]
        to_ignore = [
            (bookworm, "index:something-else"),
            (bookworm, "hello_1.0"),
            (trixie, "index:Release"),
        ]
        for collection, name in to_remove + to_ignore:
            for created_at in dts:
                # Even during test setup, we have to remove old items to
                # avoid constraint violations.
                collection.child_items.active().filter(
                    child_type=CollectionItem.Types.ARTIFACT, name=name
                ).update(
                    removed_by_user=self.playground.get_default_user(),
                    removed_at=created_at,
                )
                artifact, _ = self.playground.create_artifact()
                item = CollectionItem.objects.create_from_artifact(
                    artifact,
                    parent_collection=collection,
                    name=name,
                    data={},
                    created_by_user=self.playground.get_default_user(),
                )
                item.created_at = created_at
                item.save()

        task1 = self.create_generatesuiteindexes(
            suite_collection=suite_collection,
            generate_at=dt.datetime(2024, 12, 15, tzinfo=dt.UTC),
        )
        task1.supersede_older_items()

        expected_values: list[
            tuple[int, str, dt.datetime, dt.datetime | None]
        ] = []
        for collection, name in to_remove + to_ignore:
            expected_values += [
                (collection.id, name, dts[0], dts[1]),
                (collection.id, name, dts[1], dts[2]),
                (collection.id, name, dts[2], dts[3]),
                (collection.id, name, dts[3], None),
            ]
        self.assertQuerySetEqual(
            CollectionItem.objects.filter(
                parent_collection__in={bookworm, trixie}
            ).values_list(
                "parent_collection", "name", "created_at", "removed_at"
            ),
            expected_values,
            ordered=False,
        )

        task2 = self.create_generatesuiteindexes(
            suite_collection=suite_collection,
            generate_at=dt.datetime(2025, 3, 15, tzinfo=dt.UTC),
        )
        task2.supersede_older_items()

        expected_values = []
        for collection, name in to_remove:
            expected_values += [
                (collection.id, name, dts[0], dts[1]),
                (collection.id, name, dts[1], dts[2]),
                (collection.id, name, dts[2], task2.data.generate_at),
                (collection.id, name, dts[3], None),
            ]
        for collection, name in to_ignore:
            expected_values += [
                (collection.id, name, dts[0], dts[1]),
                (collection.id, name, dts[1], dts[2]),
                (collection.id, name, dts[2], dts[3]),
                (collection.id, name, dts[3], None),
            ]
        self.assertQuerySetEqual(
            CollectionItem.objects.filter(
                parent_collection__in={bookworm, trixie}
            ).values_list(
                "parent_collection", "name", "created_at", "removed_at"
            ),
            expected_values,
            ordered=False,
        )

        task3 = self.create_generatesuiteindexes(
            suite_collection=suite_collection,
            generate_at=dt.datetime(2025, 4, 15, tzinfo=dt.UTC),
        )
        task3.supersede_older_items()

        expected_values = []
        for collection, name in to_remove:
            expected_values += [
                (collection.id, name, dts[0], dts[1]),
                (collection.id, name, dts[1], dts[2]),
                (collection.id, name, dts[2], task2.data.generate_at),
                (collection.id, name, dts[3], task3.data.generate_at),
            ]
        for collection, name in to_ignore:
            expected_values += [
                (collection.id, name, dts[0], dts[1]),
                (collection.id, name, dts[1], dts[2]),
                (collection.id, name, dts[2], dts[3]),
                (collection.id, name, dts[3], None),
            ]
        self.assertQuerySetEqual(
            CollectionItem.objects.filter(
                parent_collection__in={bookworm, trixie}
            ).values_list(
                "parent_collection", "name", "created_at", "removed_at"
            ),
            expected_values,
            ordered=False,
        )

    def test_find_newer_item(self) -> None:
        """Find the next generation of index files."""
        bookworm = self.bookworm
        trixie = self.create_suite_collection("trixie")
        dts = [
            dt.datetime(2025, month, 1, tzinfo=dt.UTC) for month in (1, 2, 3, 4)
        ]
        to_remove = [
            (bookworm, "index:main/source/Sources.xz"),
            (bookworm, "index:main/binary-amd64/Packages.xz"),
            (bookworm, "index:Release"),
        ]
        to_ignore = [
            (bookworm, "index:something-else"),
            (bookworm, "hello_1.0"),
            (trixie, "index:Release"),
        ]
        for collection, name in to_remove + to_ignore:
            for created_at in dts:
                # Even during test setup, we have to remove old items to
                # avoid constraint violations.
                collection.child_items.active().filter(
                    child_type=CollectionItem.Types.ARTIFACT, name=name
                ).update(
                    removed_by_user=self.playground.get_default_user(),
                    removed_at=created_at,
                )
                artifact, _ = self.playground.create_artifact()
                item = CollectionItem.objects.create_from_artifact(
                    artifact,
                    parent_collection=collection,
                    name=name,
                    data={},
                    created_by_user=self.playground.get_default_user(),
                )
                item.created_at = created_at
                item.save()

        for generate_at, expected_item in (
            (
                dt.datetime(2025, 2, 15, tzinfo=dt.UTC),
                bookworm.child_items.get(
                    name="index:main/source/Sources.xz", created_at=dts[2]
                ),
            ),
            (
                dt.datetime(2025, 3, 1, tzinfo=dt.UTC),
                bookworm.child_items.get(
                    name="index:main/source/Sources.xz", created_at=dts[3]
                ),
            ),
            (dt.datetime(2025, 4, 15, tzinfo=dt.UTC), None),
        ):
            with self.subTest(generate_at=generate_at.isoformat()):
                self.task_reset_current()
                self.task_data["generate_at"] = generate_at
                self.assertEqual(self.task.find_newer_item(), expected_item)

    def test_execute_lock_error(self) -> None:
        """The task fails if its lock is already held."""
        bookworm = self.bookworm
        with (
            self.separate_connection(conflicting_alias := "conflicting-lock"),
            advisory_lock(
                (
                    LockType.GENERATE_SUITE_INDEXES,
                    bookworm.id & (2**31 - 1),
                ),
                wait=False,
                using=conflicting_alias,
            ) as acquired,
            self.assertRaisesRegex(
                LockError,
                f"Another GenerateSuiteIndexes task for "
                f"{self.task_data["suite_collection"]} is already running",
            ),
        ):
            assert acquired
            self.task.execute()

    def test_get_label(self) -> None:
        """Test get_label."""
        self.task_data["generate_at"] = dt.datetime(2025, 1, 1, tzinfo=dt.UTC)
        self.assertEqual(
            self.task.get_label(),
            f"generate indexes for {self.task_data["suite_collection"]} "
            f"at 2025-01-01 00:00:00+00:00",
        )


class GenerateSuiteIndexesUIPlaygroundTests(GenerateSuiteIndexesTestsBase):
    """Tests for :py:class:`GenerateSuiteIndexes` on UIPlayground."""

    SAMPLE_TASK_DATA = {
        "suite_collection": f"play_bookworm@{CollectionCategory.SUITE}",
        "generate_at": dt.datetime.now(dt.UTC),
    }

    scenario = scenarios.UIPlayground(set_current=True)

    def test_execute(self) -> None:
        """Test end-to-end execution."""
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        dts = [
            dt.datetime(2025, month, 1, tzinfo=dt.UTC) for month in (1, 2, 3)
        ]
        older_release = self.playground.create_repository_index("Release")
        older_release_item = self.scenario.suite.manager.add_artifact(
            older_release,
            user=self.playground.get_default_user(),
            variables={"path": "Release"},
        )
        older_release_item.created_at = dts[0]
        older_release_item.removed_by_user = self.playground.get_default_user()
        older_release_item.removed_at = dts[2]
        older_release_item.save()
        newer_release = self.playground.create_repository_index("Release")
        newer_release_item = self.scenario.suite.manager.add_artifact(
            newer_release,
            user=self.playground.get_default_user(),
            variables={"path": "Release"},
        )
        newer_release_item.created_at = dts[2]
        newer_release_item.save()

        self.task_data["suite_collection"] = suite_collection
        self.task_data["generate_at"] = dts[1]

        # Pretend that a lock for another collection is held, to make sure
        # that such a lock doesn't interfere.
        with (
            self.separate_connection(
                non_conflicting_alias := "non-conflicting-lock"
            ),
            advisory_lock(
                (
                    LockType.GENERATE_SUITE_INDEXES,
                    (self.scenario.suite.id + 1) & (2**31 - 1),
                ),
                wait=False,
                using=non_conflicting_alias,
            ) as acquired,
        ):
            assert acquired
            self.assertEqual(self.task.execute(), WorkRequest.Results.SUCCESS)

        self.assertQuerySetEqual(
            self.scenario.suite.child_items.filter(
                name__startswith="index:"
            ).values_list("name", "created_at", "removed_at"),
            [
                (f"index:{component}/source/Sources.xz", dts[1], dts[2])
                for component in self.scenario.suite.data["components"]
            ]
            + [
                (
                    f"index:{component}/binary-{architecture}/Packages.xz",
                    dts[1],
                    dts[2],
                )
                for component in self.scenario.suite.data["components"]
                for architecture in self.scenario.suite.data["architectures"]
            ]
            + [
                ("index:Release", dts[0], dts[1]),
                ("index:Release", dts[1], dts[2]),
                ("index:Release", dts[2], None),
            ],
            ordered=False,
        )

    def test_generate_sources(self) -> None:
        """``generate_sources`` generates a ``Sources`` file."""
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        task = self.create_generatesuiteindexes(
            suite_collection=suite_collection
        )
        temp_path = self.create_temporary_directory()

        task.generate_sources(temp_path=temp_path, component="main")

        dpkg = task.suite.collection.manager.lookup("source:dpkg")
        assert dpkg is not None
        assert dpkg.artifact is not None
        dpkg_dsc = dpkg.artifact.fileinartifact_set.get(
            path="dpkg_1.21.22.dsc"
        ).file
        dpkg_orig = dpkg.artifact.fileinartifact_set.get(
            path="dpkg_1.21.22.tar.gz"
        ).file
        hello = task.suite.collection.manager.lookup("source:hello")
        assert hello is not None
        assert hello.artifact is not None
        hello_dsc = hello.artifact.fileinartifact_set.get(
            path="hello_1.0-1.dsc"
        ).file
        hello_debian = hello.artifact.fileinartifact_set.get(
            path="hello_1.0-1.debian.tar.xz"
        ).file
        hello_orig = hello.artifact.fileinartifact_set.get(
            path="hello_1.0.orig.tar.gz"
        ).file
        expected_sources = dedent(
            f"""\
            Package: dpkg
            Binary: dpkg, dpkg-doc, libdpkg-perl
            Version: 1.21.22
            Maintainer: Example Maintainer <example@example.org>
            Architecture: any
            Standards-Version: 4.3.0
            Format: 3.0 (quilt)
            Checksums-Sha256:
             {dpkg_dsc.sha256.hex()} {dpkg_dsc.size} dpkg_1.21.22.dsc
             {dpkg_orig.sha256.hex()} {dpkg_orig.size} dpkg_1.21.22.tar.gz
            Homepage: http://www.gnu.org/software/dpkg/
            Package-List:{" "}
             dpkg deb devel optional arch=any
             dpkg-doc deb devel optional arch=any
             libdpkg-perl deb devel optional arch=any
            Directory: pool/main/d/dpkg
            Priority: source
            Section: devel

            Package: hello
            Binary: hello
            Version: 1.0-1
            Maintainer: Example Maintainer <example@example.org>
            Architecture: any
            Standards-Version: 4.3.0
            Format: 3.0 (quilt)
            Checksums-Sha256:
             {hello_debian.sha256.hex()} {hello_debian.size} hello_1.0-1.debian.tar.xz
             {hello_dsc.sha256.hex()} {hello_dsc.size} hello_1.0-1.dsc
             {hello_orig.sha256.hex()} {hello_orig.size} hello_1.0.orig.tar.gz
            Homepage: http://www.gnu.org/software/hello/
            Package-List:{" "}
             hello deb devel optional arch=any
            Directory: pool/main/h/hello
            Priority: source
            Section: devel

            """  # noqa: E501
        )

        self.assertIsNone(
            task.suite.collection.manager.lookup("index:main/source/Sources")
        )
        sources_xz_item = task.suite.collection.manager.lookup(
            "index:main/source/Sources.xz"
        )
        assert sources_xz_item is not None
        self.assertEqual(sources_xz_item.created_at, task.data.generate_at)
        assert sources_xz_item.artifact is not None
        sources_xz_file = sources_xz_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(
            sources_xz_file
        )
        with file_backend.get_stream(sources_xz_file) as compressed:
            got_sources = lzma.decompress(compressed.read())
            self.assertEqual(got_sources.decode(), expected_sources)
        self.assertEqual(
            task.index_info,
            {
                PurePath("main", "source", "Sources"): IndexInfo(
                    size=len(got_sources),
                    sha256=hashlib.sha256(got_sources).hexdigest(),
                ),
                PurePath("main", "source", "Sources.xz"): IndexInfo(
                    size=sources_xz_file.size,
                    sha256=sources_xz_file.sha256.hex(),
                    artifact=sources_xz_item.artifact,
                ),
            },
        )

    def test_generate_packages(self) -> None:
        """``generate_packages`` generates a ``Packages`` file."""
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        task = self.create_generatesuiteindexes(
            suite_collection=suite_collection
        )
        temp_path = self.create_temporary_directory()

        task.generate_packages(
            temp_path=temp_path, component="main", architecture="amd64"
        )

        dpkg_amd64 = task.suite.collection.manager.lookup("binary:dpkg_amd64")
        assert dpkg_amd64 is not None
        assert dpkg_amd64.artifact is not None
        dpkg_amd64_deb = dpkg_amd64.artifact.fileinartifact_set.get(
            path="dpkg_1.21.22_amd64.deb"
        ).file
        hello_amd64 = task.suite.collection.manager.lookup("binary:hello_amd64")
        assert hello_amd64 is not None
        assert hello_amd64.artifact is not None
        hello_amd64_deb = hello_amd64.artifact.fileinartifact_set.get(
            path="hello_1.0-1_amd64.deb"
        ).file
        expected_packages = dedent(
            f"""\
            Package: dpkg
            Version: 1.21.22
            Maintainer: Example Maintainer <example@example.org>
            Architecture: amd64
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/d/dpkg/dpkg_1.21.22_amd64.deb
            Size: {dpkg_amd64_deb.size}
            SHA256: {dpkg_amd64_deb.sha256.hex()}

            Package: hello
            Version: 1.0-1
            Maintainer: Example Maintainer <example@example.org>
            Architecture: amd64
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/h/hello/hello_1.0-1_amd64.deb
            Size: {hello_amd64_deb.size}
            SHA256: {hello_amd64_deb.sha256.hex()}

            """
        )

        self.assertIsNone(
            task.suite.collection.manager.lookup(
                "index:main/binary-amd64/Packages"
            )
        )
        packages_xz_item = task.suite.collection.manager.lookup(
            "index:main/binary-amd64/Packages.xz"
        )
        assert packages_xz_item is not None
        self.assertEqual(packages_xz_item.created_at, task.data.generate_at)
        assert packages_xz_item.artifact is not None
        packages_xz_file = packages_xz_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(
            packages_xz_file
        )
        with file_backend.get_stream(packages_xz_file) as compressed:
            got_packages = lzma.decompress(compressed.read())
            self.assertEqual(got_packages.decode(), expected_packages)
        self.assertEqual(
            task.index_info,
            {
                PurePath("main", "binary-amd64", "Packages"): IndexInfo(
                    size=len(got_packages),
                    sha256=hashlib.sha256(got_packages).hexdigest(),
                ),
                PurePath("main", "binary-amd64", "Packages.xz"): IndexInfo(
                    size=packages_xz_file.size,
                    sha256=packages_xz_file.sha256.hex(),
                    artifact=packages_xz_item.artifact,
                ),
            },
        )

    def test_generate_packages_duplicate_architecture_all(self) -> None:
        """The ``duplicate_architecture_all`` flag works."""
        self.scenario.suite.data["duplicate_architecture_all"] = True
        self.scenario.suite.save()
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        task = self.create_generatesuiteindexes(
            suite_collection=suite_collection
        )
        temp_path = self.create_temporary_directory()

        task.generate_packages(
            temp_path=temp_path, component="main", architecture="amd64"
        )

        debs: dict[str, File] = {}
        for pkg in (
            "dpkg_amd64",
            "dpkg-doc_all",
            "libdpkg-perl_all",
            "hello_amd64",
        ):
            item = task.suite.collection.manager.lookup(f"binary:{pkg}")
            assert item is not None
            assert item.artifact is not None
            debs[pkg] = item.artifact.fileinartifact_set.get().file

        expected_packages = dedent(
            f"""\
            Package: dpkg
            Version: 1.21.22
            Maintainer: Example Maintainer <example@example.org>
            Architecture: amd64
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/d/dpkg/dpkg_1.21.22_amd64.deb
            Size: {debs["dpkg_amd64"].size}
            SHA256: {debs["dpkg_amd64"].sha256.hex()}

            Package: dpkg-doc
            Source: dpkg
            Version: 1.21.22
            Maintainer: Example Maintainer <example@example.org>
            Architecture: all
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/d/dpkg/dpkg-doc_1.21.22_all.deb
            Size: {debs["dpkg-doc_all"].size}
            SHA256: {debs["dpkg-doc_all"].sha256.hex()}

            Package: libdpkg-perl
            Source: dpkg
            Version: 1.21.22
            Maintainer: Example Maintainer <example@example.org>
            Architecture: all
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/d/dpkg/libdpkg-perl_1.21.22_all.deb
            Size: {debs["libdpkg-perl_all"].size}
            SHA256: {debs["libdpkg-perl_all"].sha256.hex()}

            Package: hello
            Version: 1.0-1
            Maintainer: Example Maintainer <example@example.org>
            Architecture: amd64
            Description: Example description
            Section: devel
            Priority: optional
            Filename: pool/main/h/hello/hello_1.0-1_amd64.deb
            Size: {debs["hello_amd64"].size}
            SHA256: {debs["hello_amd64"].sha256.hex()}

            """
        )

        self.assertIsNone(
            task.suite.collection.manager.lookup(
                "index:main/binary-amd64/Packages"
            )
        )
        packages_xz_item = task.suite.collection.manager.lookup(
            "index:main/binary-amd64/Packages.xz"
        )
        assert packages_xz_item is not None
        assert packages_xz_item.artifact is not None
        packages_xz_file = packages_xz_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(
            packages_xz_file
        )
        with file_backend.get_stream(packages_xz_file) as compressed:
            got_packages = lzma.decompress(compressed.read())
            self.assertEqual(got_packages.decode(), expected_packages)
        self.assertEqual(
            task.index_info,
            {
                PurePath("main", "binary-amd64", "Packages"): IndexInfo(
                    size=len(got_packages),
                    sha256=hashlib.sha256(got_packages).hexdigest(),
                ),
                PurePath("main", "binary-amd64", "Packages.xz"): IndexInfo(
                    size=packages_xz_file.size,
                    sha256=packages_xz_file.sha256.hex(),
                    artifact=packages_xz_item.artifact,
                ),
            },
        )

    def test_generate_release(self) -> None:
        """``generate_release`` generates a ``Release`` file."""
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        task = self.create_generatesuiteindexes(
            suite_collection=suite_collection
        )
        temp_path = self.create_temporary_directory()
        sources_xz_artifact = self.playground.create_repository_index(
            "Sources.xz"
        )
        sources_xz = sources_xz_artifact.files.get()
        packages_xz_artifact = self.playground.create_repository_index(
            "Packages.xz"
        )
        packages_xz = packages_xz_artifact.files.get()
        task.index_info = {
            PurePath("main", "source", "Sources"): IndexInfo(
                size=1, sha256="sources-sha256"
            ),
            PurePath("main", "source", "Sources.xz"): IndexInfo(
                size=sources_xz.size,
                sha256=sources_xz.sha256.hex(),
                artifact=sources_xz_artifact,
            ),
            PurePath("main", "binary-amd64", "Packages"): IndexInfo(
                size=2, sha256="packages-sha256"
            ),
            PurePath("main", "binary-amd64", "Packages.xz"): IndexInfo(
                size=packages_xz.size,
                sha256=packages_xz.sha256.hex(),
                artifact=packages_xz_artifact,
            ),
        }

        task.generate_release(temp_path=temp_path)

        expected_date = task.data.generate_at.strftime(
            "%a, %d %b %Y %H:%M:%S UTC"
        )
        expected_release = dedent(
            f"""\
            Suite: stable
            Codename: bookworm
            Date: {expected_date}
            Acquire-By-Hash: yes
            Architectures: {" ".join(task.architectures)}
            Components: {" ".join(task.components)}
            SHA256:
             packages-sha256 {2:16} main/binary-amd64/Packages
             {packages_xz.sha256.hex()} {packages_xz.size:16} main/binary-amd64/Packages.xz
             sources-sha256 {1:16} main/source/Sources
             {sources_xz.sha256.hex()} {sources_xz.size:16} main/source/Sources.xz
            """  # noqa: E501
        )

        release_item = task.suite.collection.manager.lookup("index:Release")
        assert release_item is not None
        self.assertEqual(release_item.created_at, task.data.generate_at)
        assert release_item.artifact is not None
        self.assertEqual(
            release_item.artifact.fileinartifact_set.get().content_type,
            "text/plain; charset=utf-8",
        )
        release_file = release_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(release_file)
        with file_backend.get_stream(release_file) as f:
            got_release = f.read().decode()
            self.assertEqual(got_release, expected_release)
        self.assertQuerySetEqual(
            Artifact.objects.filter(
                targeted_by__artifact=release_item.artifact
            ),
            [sources_xz_artifact, packages_xz_artifact],
            ordered=False,
        )

    def test_generate_release_field_ordering(self) -> None:
        """``generate_release`` orders ``Release`` fields appropriately."""
        self.scenario.suite.data["release_fields"].update(
            {
                "Version": "12.0",
                "NotAutomatic": True,
                "ButAutomaticUpgrades": False,
                "Acquire-By-Hash": "yes",
                "Custom-Field": "foo",
            }
        )
        self.scenario.suite.data["duplicate_architecture_all"] = True
        self.scenario.suite.save()
        suite_collection = f"play_bookworm@{CollectionCategory.SUITE}"
        task = self.create_generatesuiteindexes(
            suite_collection=suite_collection
        )
        temp_path = self.create_temporary_directory()
        sources_xz_artifact = self.playground.create_repository_index(
            "Sources.xz"
        )
        sources_xz = sources_xz_artifact.files.get()
        task.index_info = {
            PurePath("main", "source", "Sources"): IndexInfo(
                size=1, sha256="sources-sha256"
            ),
            PurePath("main", "source", "Sources.xz"): IndexInfo(
                size=sources_xz.size,
                sha256=sources_xz.sha256.hex(),
                artifact=sources_xz_artifact,
            ),
        }

        task.generate_release(temp_path=temp_path)

        expected_date = task.data.generate_at.strftime(
            "%a, %d %b %Y %H:%M:%S UTC"
        )
        expected_release = dedent(
            f"""\
            Suite: stable
            Version: 12.0
            Codename: bookworm
            Date: {expected_date}
            NotAutomatic: yes
            Acquire-By-Hash: yes
            No-Support-for-Architecture-all: Packages
            Architectures: {" ".join(task.architectures)}
            Components: {" ".join(task.components)}
            Custom-Field: foo
            SHA256:
             sources-sha256 {1:16} main/source/Sources
             {sources_xz.sha256.hex()} {sources_xz.size:16} main/source/Sources.xz
            """  # noqa: E501
        )

        release_item = task.suite.collection.manager.lookup("index:Release")
        assert release_item is not None
        assert release_item.artifact is not None
        release_file = release_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(release_file)
        with file_backend.get_stream(release_file) as f:
            got_release = f.read().decode()
            self.assertEqual(got_release, expected_release)
        self.assertQuerySetEqual(
            Artifact.objects.filter(
                targeted_by__artifact=release_item.artifact
            ),
            [sources_xz_artifact],
        )

    def test_override_acquire_by_hash(self) -> None:
        """``generate_release`` lets ``Acquire-By-Hash: yes`` be overridden."""
        self.scenario.suite.data["release_fields"]["Acquire-By-Hash"] = False
        self.scenario.suite.save()

        temp_path = self.create_temporary_directory()
        self.task.generate_release(temp_path=temp_path)

        release_item = self.task.suite.collection.manager.lookup(
            "index:Release"
        )
        assert release_item is not None
        assert release_item.artifact is not None
        release_file = release_item.artifact.files.get()
        file_backend = self.scenario.scope.download_file_backend(release_file)
        with file_backend.get_stream(release_file) as f:
            got_release = f.read().decode()
            self.assertNotIn("Acquire-By-Hash:", got_release)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Debusine client."""

from debusine.artifacts.local_artifact import (
    AutopkgtestArtifact,
    BinaryPackage,
    BinaryPackages,
    BlhcArtifact,
    DebDiffArtifact,
    DebianSystemTarballArtifact,
    DebugSymbols,
    LintianArtifact,
    LocalArtifact,
    PackageBuildLog,
    PiupartsArtifact,
    RepositoryIndex,
    SigningInputArtifact,
    SigningOutputArtifact,
    SourcePackage,
    Upload,
    WorkRequestDebugLogs,
)

__all__ = [
    "AutopkgtestArtifact",
    "BinaryPackage",
    "BinaryPackages",
    "BlhcArtifact",
    "DebDiffArtifact",
    "DebianSystemTarballArtifact",
    "DebugSymbols",
    "LintianArtifact",
    "LocalArtifact",
    "PackageBuildLog",
    "PiupartsArtifact",
    "RepositoryIndex",
    "SigningInputArtifact",
    "SigningOutputArtifact",
    "SourcePackage",
    "Upload",
    "WorkRequestDebugLogs",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Test Django REST Framework authentication infrastructure for Debusine."""

from typing import ClassVar

from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory
from rest_framework.request import Request

from debusine.db.models import Token, User
from debusine.server.views.auth import DebusineTokenAuthentication
from debusine.test.django import TestCase


class DebusineTokenAuthenticationTests(TestCase):
    """Test DebusineTokenAuthentication."""

    token: ClassVar[Token]

    def authenticate(
        self, token: Token | str | None = None
    ) -> tuple[User | AnonymousUser | None, Token | None] | None:
        """Run the authentication method."""
        django_request = RequestFactory().get("/")
        setattr(django_request, "_debusine_token", token)
        request = Request(django_request)
        return DebusineTokenAuthentication().authenticate(request)

    def test_worker_token(self) -> None:
        """Check authenticating a worker token."""
        token = self.playground.create_worker_token()
        self.assertEqual(self.authenticate(token), (AnonymousUser(), token))

    def test_user_token(self) -> None:
        """Check authenticating a user token."""
        token = self.playground.create_user_token()
        self.assertEqual(
            self.authenticate(token),
            (self.playground.get_default_user(), token),
        )

    def test_no_token(self) -> None:
        """Request without a token skip authentication."""
        self.assertIsNone(self.authenticate())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Worker client: connects to debusine server.

Overview
--------
* Registration (needed only once per worker): If the worker doesn't have a
  token: it will generate it and register with the server
  (HTTP POST to ``/api/1.0/worker/register``)
* The client will use this token to connect to the server (WebSocket to
  ``/api/ws/1.0/worker/connect``)

Flow
----

  #. The worker is executed and chooses ``~/.config/debusine/worker``
     (if it exists) or ``/etc/debusine/worker``. It reads the file
     ``config.ini`` from the directory and if it already exists the file
     ``token``.
  #. If there isn't a token the worker generates one (using
     :py:func:`secrets.token_hex`) and registers it to the Debusine server
     via HTTP POST to `/api/1.0/worker/register` sending the generated token
     and the worker's FQDN. The token is saved to the ``token`` file in the
     chosen config directory.
  #. The server will create a new Token and Worker in the DB via the models.
     They wouldn't be used until manual validation.
  #. The client can then connect using WebSockets to
     ``/api/ws/1.0/worker/connect`` and wait for commands to execute.

Objects documentation
---------------------

"""

import abc
import asyncio
import functools
import logging
import secrets
import shutil
import signal
import socket
import time as tm
import types
from pathlib import Path
from textwrap import dedent
from threading import Lock
from typing import Any, Literal, NoReturn, Self, override

import aiohttp
import tenacity
from aiohttp.web_exceptions import HTTPCreated, HTTPNoContent, HTTPOk
from pydantic import ValidationError
from systemd.journal import JournalHandler

from debusine.artifacts.models import WorkRequestResults
from debusine.client.debusine import Debusine
from debusine.client.models import WorkRequestResponse
from debusine.tasks import (
    analyze_external_worker_tasks,
    get_provided_external_worker_tags,
)
from debusine.tasks.executors import (
    analyze_worker_all_executors,
    executor_provided_tags,
)
from debusine.tasks.models import OutputData, OutputDataError, WorkerType
from debusine.utils import debusine_version
from debusine.worker.config import ConfigHandler
from debusine.worker.debusine_async_http_client import DebusineAsyncHttpClient
from debusine.worker.system_information import (
    system_metadata,
    system_provided_tags,
)
from debusine.worker.task_runner import TaskRunner, TaskRunnerError

logger = logging.getLogger("debusine.worker")


class Event:
    """Base class for worker events."""

    def __init__(self, name: str) -> None:
        self.name = name
        self.perf_ns = tm.perf_counter_ns()

    @override
    def __str__(self) -> str:
        return self.name


class ShutdownEvent(Event, abc.ABC):
    """Shut down the worker."""

    def __init__(self, message: str) -> None:
        super().__init__("shutdown")
        self.message = message

    @override
    def __str__(self) -> str:
        return f"{self.name}: {self.message}"

    @abc.abstractmethod
    def get_exit_code(self) -> int:
        """Return the exit code to use."""


class ServerDisconnectShutdownEvent(ShutdownEvent):
    """Shut down because the server disconnected the event stream."""

    def __init__(self) -> None:
        super().__init__("server notification connection closed")

    @override
    def get_exit_code(self) -> int:
        return 1


class ReportTaskResultFailedShutdownEvent(ShutdownEvent):
    """Shut down after failing to report the result of a task."""

    def __init__(self) -> None:
        super().__init__("cannot send results to server")

    @override
    def get_exit_code(self) -> int:
        return 2


class SignalReceivedShutdownEvent(ShutdownEvent):
    """Shutdown after receiving a signal."""

    def __init__(self, sig: signal.Signals) -> None:
        super().__init__(f"signal {sig} received")
        self.sig = sig

    @override
    def get_exit_code(self) -> int:
        return 128 + self.sig


class ServerEvent(Event):
    """Server-triggered event."""

    def __init__(self, name: str) -> None:
        super().__init__(name)

    @classmethod
    def from_message(cls, payload: dict[str, Any]) -> "ServerEvent":
        """Instantiate a server event from a server message."""
        match payload.get("text"):
            case "connected":
                return ConnectedEvent()
            case "request_dynamic_metadata":
                return RequestDynamicMetadataEvent()
            case "work_request_available":
                return WorkRequestAvailableEvent()
            case None:
                raise ValueError(f"{payload!r}: message payload without 'text'")
            case _:
                raise NotImplementedError(
                    f"{payload!r}: unsupported server event"
                )


class ConnectedEvent(ServerEvent):
    """Connection to server established."""

    def __init__(self) -> None:
        super().__init__("connected")


class RequestDynamicMetadataEvent(ServerEvent):
    """Server requested dynamic metadata."""

    def __init__(self) -> None:
        super().__init__("request_dynamic_metadata")


class WorkRequestAvailableEvent(ServerEvent):
    """Server notifies that there is a work request available."""

    def __init__(self) -> None:
        super().__init__("work_request_available")


class QuitMainLoopEvent(Event):
    """
    Exit the main loop.

    This is mostly used by tests to exit the main loop waiting for pending
    messages.
    """

    def __init__(self) -> None:
        super().__init__("quit_main_loop")


class Worker:
    """Worker class: waits for commands from the debusine server."""

    DEFAULT_LOG_LEVEL = logging.INFO
    AFTER_WORK_REQUEST_HOOK_DIR = Path(
        "/etc/debusine/worker/hooks/after-work-request.d"
    )

    def __init__(
        self,
        *,
        log_level: str | None = None,
        worker_type: Literal[
            WorkerType.EXTERNAL, WorkerType.SIGNING
        ] = WorkerType.EXTERNAL,
        config: ConfigHandler | None = None,
        working_directory: str | None = None,
    ) -> None:
        """
        Initialize Worker.

        :param log_level: minimum level of the logs being saved. If None uses
          settings from config.ini or DEFAULT_LOG_LEVEL.
        :param config: ConfigHandler to use (or creates a default one)
        """
        self._original_log_level = log_level
        self._worker_type = worker_type

        if config is None:
            if worker_type == WorkerType.SIGNING:
                self._config = ConfigHandler(
                    directories=[
                        str(Path.home() / ".config/debusine/signing"),
                        "/etc/debusine/signing",
                    ],
                    require_https=True,
                )
            else:
                self._config = ConfigHandler()
        else:
            self._config = config

        self._setup_logging()

        self._setup_workdir(working_directory)

        self._config.validate_config_or_fail()

        self._aiohttp_client_session: aiohttp.ClientSession | None = None

        self._async_http_client = DebusineAsyncHttpClient(self._config)

        # Main event queue
        self.event_queue: asyncio.Queue[Event] = asyncio.Queue()

        # Used by ThreadPoolExecutor to submit tasks to the main event loop
        self._main_event_loop: asyncio.AbstractEventLoop | None = None

        # Lock to protect from concurrent work request execution
        self._task_lock: Lock = Lock()

    async def __aenter__(self) -> Self:
        """Async context manager support."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Release worker resources."""
        await self._async_http_client.close()

    def _setup_workdir(self, working_directory: str | None) -> None:
        if working_directory:
            self._workdir = Path(working_directory)
        else:
            self._workdir = self._config.working_directory

        # Create the working directory if needed
        try:
            self._workdir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            self._fail(
                f"could not create working directory '{self._workdir}': {e}"
            )

    def clean_workdir(self) -> None:
        """Clean the working directory from interrupted work requests."""
        for entry in self._workdir.iterdir():
            if entry.is_dir() and entry.name.startswith("WR-"):
                try:
                    shutil.rmtree(entry)
                except OSError as e:
                    logger.warning("Failed to remove '%s': %s", entry, str(e))

    def _setup_logging(self) -> None:
        """
        Set logging configuration.

        Use the parameters passed to Worker.__init__() and self._config
        configuration.
        """
        log_level = (
            self._original_log_level
            or self._config.log_level
            or logging.getLevelName(self.DEFAULT_LOG_LEVEL)
        )

        logging.basicConfig(
            format='%(message)s',
            level=log_level,
            handlers=[JournalHandler()],
            force=True,
        )

    def _create_task_signal_int_term_handler(
        self, signum: signal.Signals
    ) -> asyncio.Task[None]:
        return asyncio.create_task(self._signal_int_term_handler(signum))

    async def main(self) -> None:
        """Run the worker."""
        self.clean_workdir()
        logger.info(
            "Debusine worker is starting. Version: %s", debusine_version()
        )
        for signum in [signal.SIGINT, signal.SIGTERM]:
            asyncio.get_running_loop().add_signal_handler(
                signum,
                functools.partial(
                    self._create_task_signal_int_term_handler, signum
                ),
            )

        await self.require_token()
        if event := await self.process_notifications():
            raise SystemExit(event.get_exit_code())
        else:
            logger.warning("Unknown shutdown cause")
            raise SystemExit(4)

    async def process_notifications(self) -> ShutdownEvent | None:
        """Process server notifications."""

        class Shutdown(Exception):
            """Shutdown has been requested."""

        shutdown_event: ShutdownEvent | None = None

        try:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(self.handle_server_notifications())
                while True:
                    event = await self.event_queue.get()
                    logger.debug("Worker event: %s", event)
                    match event:
                        case ShutdownEvent():
                            # Raise an exception instead of breaking out of the
                            # while, so that tasks in tg are cancelled
                            shutdown_event = event
                            raise Shutdown(str(event))
                        case ConnectedEvent():
                            logger.debug("Received the 'connected' message")
                        case RequestDynamicMetadataEvent():
                            tg.create_task(self._send_dynamic_metadata())
                        case WorkRequestAvailableEvent():
                            tg.create_task(
                                self._request_and_execute_work_request()
                            )
                        case QuitMainLoopEvent():
                            break
                        case _:
                            logger.warning(
                                "Ignoring unsupported event: %s", event
                            )
        except* Shutdown as exc:
            logger.info("Shutdown requested: %s", exc.exceptions[0])

        return shutdown_event

    async def require_token(self) -> None:
        """Handle registration if needed."""
        if not self._config.token:
            result = await self._register()
            if result is False:
                self._fail('Exiting...')

    async def _signal_int_term_handler(self, signum: signal.Signals) -> None:
        self.log_forced_exit(signum)
        await self.event_queue.put(SignalReceivedShutdownEvent(signum))

    @staticmethod
    def _log_server_error(status_code: int, body: str) -> None:
        logger.error(
            'Could not register. Server HTTP status code: %s Body: %s\n',
            status_code,
            body,
        )

    async def _register(self) -> bool:
        """
        Create a token, registers it to debusine, saves it locally.

        The worker will not receive any tasks until the debusine admin
        has approved the token.
        """
        token = secrets.token_hex(32)

        register_path = '/1.0/worker/register/'

        data = {
            "token": token,
            "fqdn": socket.getfqdn(),
            "worker_type": self._worker_type,
        }

        try:
            response = await self._async_http_client.post(
                register_path, json=data, token=self._config.activation_token
            )
        except aiohttp.client_exceptions.ClientResponseError as err:
            self._log_server_error(err.status, 'Not available')
            return False
        except aiohttp.client_exceptions.ClientConnectorError as err:
            logger.error(  # noqa: G200
                'Could not register. Server unreachable: %s', err
            )
            return False

        status, body = (
            response.status,
            (await response.content.read()).decode('utf-8'),
        )

        if status == HTTPCreated.status_code:
            self._config.write_token(token)
            return True
        else:
            self._log_server_error(status, body)
            return False

    @staticmethod
    async def _asyncio_sleep(delay: float) -> None:
        """Sleep asynchronously (mocked in tests)."""
        await asyncio.sleep(delay)

    async def handle_server_notifications(self) -> None:
        """Read incoming server notifications and generate worker events."""
        client = Debusine(
            self._config.api_url,
            self._config.token,
            timeout=self._config.http_timeout,
            logger=logger,
            scope=None,
        )
        async with client.server_notifications(
            endpoint="1.0/worker/connect/"
        ) as server_notifications:
            async for payload in server_notifications.messages():
                try:
                    event = ServerEvent.from_message(payload)
                except Exception as exc:
                    logger.warning(
                        "Ignoring unsupported or malformed server message: %s",
                        exc,
                    )
                    continue
                await self.event_queue.put(event)
        await self.event_queue.put(ServerDisconnectShutdownEvent())

    @staticmethod
    def _fail(message: str) -> NoReturn:
        logger.fatal(message)
        raise SystemExit(3)

    async def _request_and_execute_work_request(self) -> None:
        # Grab the task lock to make sure that the previous work request
        # completed fully. We can be notified before the end because the
        # server notifies via websocket while we are still wrapping up
        # the task execution.
        logger.info("Requesting a work request to execute")
        retries = 0
        while not self._task_lock.acquire(blocking=False):
            await self._asyncio_sleep(1)
            retries += 1
            if retries >= 60:
                logger.error(
                    "Worker is busy and can't execute a new work request"
                )
                return
        logger.debug(
            "Waited %s seconds for the end of the former task", retries
        )

        work_request = await self._request_work_request()
        if work_request:
            logger.info(
                "Executing work request %d (%s)",
                work_request.id,
                work_request.task_name,
            )
            async with TaskRunner(
                self._config, self._worker_type, work_request, self._workdir
            ) as runner:
                try:
                    await runner.setup_task()
                    result = await runner.execute()
                except TaskRunnerError as exc:
                    logger.error(exc.message, *exc.args)
                    await self.report_task_result(
                        runner,
                        WorkRequestResults.ERROR,
                        OutputData(
                            errors=[
                                OutputDataError(
                                    message=exc.message % exc.args,
                                    code=exc.code,
                                )
                            ]
                        ),
                    )
                else:
                    # Send the task results
                    output_data = OutputData(
                        runtime_statistics=runner.statistics
                    )
                    await self.report_task_result(runner, result, output_data)

            logger.debug("Finished executing work request %d", work_request.id)
        else:
            logger.info('No work request available')
            self._task_lock.release()

    async def report_task_result(
        self,
        task_runner: TaskRunner,
        result: WorkRequestResults,
        output_data: OutputData,
    ) -> None:
        try:
            if task_runner.task and task_runner.task.aborted:
                logger.info("Task: %s has been aborted", task_runner.task.name)
                # No need to notify debusine-server
                return

            try:
                await asyncio.to_thread(
                    task_runner.client.work_request_completed_update,
                    task_runner.work_request.id,
                    result,
                    output_data,
                )
                # This is asynchronous on the server side: the server may
                # not actually have marked the work request as completed
                # yet.  However, this doesn't matter much, as the worker
                # won't do anything other than cleanup until the server
                # assigns it something else to do.
                logger.info(
                    'Task: %i completed (%s)',
                    task_runner.work_request.id,
                    result,
                )
            except Exception:
                # Log this, but leave the work request running.  The server
                # will retry it when this worker next manages to connect and
                # request a new work request to run.
                logger.exception(
                    "Cannot reach server to report work request completed. "
                    "Exiting."
                )
                # Shutdown so that the worker will try to reconnect.
                await self.event_queue.put(
                    ReportTaskResultFailedShutdownEvent()
                )
        finally:
            await self._run_after_work_request_hook()
            self._reset_state()

    async def _request_work_request(self) -> WorkRequestResponse | None:
        """Request a work request and returns it."""

        @tenacity.retry(
            sleep=self._asyncio_sleep,
            wait=tenacity.wait_random_exponential(
                multiplier=5, min=5, max=60 * 5
            ),
            retry=tenacity.retry_if_exception_type(
                aiohttp.client_exceptions.ClientError
            )
            | tenacity.retry_if_exception_type(ConnectionError),
        )
        async def inner() -> WorkRequestResponse | None:
            work_request = await self._async_http_client.post(
                '/1.0/work-request/get-next-for-worker/',
                json={},
            )
            if work_request.status == HTTPOk.status_code:
                try:
                    work_request_obj = WorkRequestResponse.model_validate_json(
                        await work_request.text()
                    )
                except ValidationError as exc:
                    logger.warning(  # noqa: G200
                        'Invalid WorkRequest received from'
                        ' /get-next-for-worker/: %s',
                        exc,
                    )
                    return None

                return work_request_obj
            else:
                work_request.raise_for_status()
                return None

        return await inner()

    def _get_provided_tags(self) -> set[str]:
        """Return a set with tags provided by this worker."""
        tags: set[str] = set()
        tags |= system_provided_tags(self._worker_type)
        tags |= executor_provided_tags()
        tags |= get_provided_external_worker_tags()
        return tags

    def _get_worker_metadata(self) -> dict[str, Any]:
        """Compute the worker metadata."""
        return {
            **system_metadata(self._worker_type),
            **analyze_worker_all_executors(),
            **analyze_external_worker_tasks(),
            "provided_tags": sorted(self._get_provided_tags()),
        }

    async def _send_dynamic_metadata(self) -> None:
        dynamic_metadata_path = '/1.0/worker/dynamic-metadata/'

        metadata = self._get_worker_metadata()

        logger.info('Send dynamic metadata')
        logger.debug('Dynamic metadata: %s', metadata)
        response = await self._async_http_client.put(
            dynamic_metadata_path, json=metadata
        )

        if response.status != HTTPNoContent.status_code:
            logger.error(
                'Failed to send dynamic_metadata: (response: %d): %s',
                response.status,
                await response.text(),
            )

    def _reset_state(self) -> None:
        """Reset worker to a clean state, ready for the next work request."""
        if self._task_lock.locked():
            # Release the task lock to allow the worker to process the next
            # work request
            self._task_lock.release()

    async def _run_after_work_request_hook(self) -> None:
        """Run any sysadmin configured post-work-request hooks."""
        hook_dir = self.AFTER_WORK_REQUEST_HOOK_DIR
        if not hook_dir.exists() or not any(hook_dir.iterdir()):
            logger.debug("No hooks in %r, skipping execution", str(hook_dir))
            return

        p = await asyncio.create_subprocess_exec(
            "run-parts",
            str(hook_dir),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await p.communicate()
        if p.returncode != 0:
            logger.warning(
                dedent(
                    """\
                    Hooks failed to execute in %r .
                    stdout:
                    %s
                    stderr:
                    %s
                    """
                ).rstrip("\n"),
                str(hook_dir),
                stdout.decode("utf-8", errors="replace"),
                stderr.decode("utf-8", errors="replace"),
            )

    @staticmethod
    def log_forced_exit(signum: signal.Signals) -> None:
        """Log a forced exit."""
        logger.info('Terminated with signal %s', signum.name)



# Generated by Django 4.2.28 on 2026-06-15 15:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0062_collectiontokengrant'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtemplate',
            name='restricted',
            field=models.BooleanField(
                default=False,
                help_text='If true, only workspace owners can start this workflow, not contributors',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine workflows views."""

from typing import cast, override

from django.db.models import QuerySet

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.db.models.work_requests import WorkRequestQuerySet
from debusine.web.views.base import ListViewBase, WorkspaceView
from debusine.web.views.places import Place
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import WorkflowTable


class WorkflowsListView(
    WorkspaceView, TableMixin[WorkRequest], ListViewBase[WorkRequest]
):
    """List workflows."""

    model = WorkRequest
    template_name = "web/workflow-list.html"
    table_class = WorkflowTable
    context_object_name = "workflow_list"
    paginate_by = 50

    @override
    def get_place(self) -> Place:
        return context.require_workspace().ui(self.request).place_workflow_list

    @override
    def get_queryset(self) -> QuerySet[WorkRequest]:
        """Filter workflows by current workspace."""
        queryset = (
            cast(WorkRequestQuerySet[WorkRequest], super().get_queryset())
            .in_current_workspace()
            .can_display(context.pc)
        )
        queryset = queryset.filter(
            task_type=TaskTypes.WORKFLOW, parent__isnull=True
        )
        return queryset



"""
The settings in this file are tailored overrides for running in production.

When running in production, selected.py should point to this file.
"""

# PostgreSQL should be used in production
try:
    from debusine.signing.settings.db_postgresql import DATABASES  # noqa: F401
except (ModuleNotFoundError, PermissionError):
    from debusine.signing.settings.db_postgresql_default import (  # noqa: F401
        DATABASES,
    )

# Use paths from the package
from debusine.signing.settings.pkg_paths import *  # noqa: F401, F403



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Authentication middleware that authenticates using signon Providers.

It adds a `request.signon` member that is a Signon object, providing an entry
point for managing externally authenticated identities.
"""

from collections.abc import Callable
from typing import Protocol, cast, runtime_checkable

import django.http
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.utils.module_loading import import_string

from debusine.server.signon.signon import Signon


@runtime_checkable
class RequestSignonProtocol(Protocol):
    """A Django request that was processed by :py:class:`SignonMiddleware`."""

    signon: Signon


class SignonMiddleware:
    """Authenticate via external signon providers."""

    signon_class: type[Signon]

    def __init__(
        self,
        get_response: Callable[
            [django.http.HttpRequest], django.http.HttpResponse
        ],
    ) -> None:
        """Middleware API entry point."""
        self.providers = getattr(settings, "SIGNON_PROVIDERS", ())
        if not self.providers:
            raise MiddlewareNotUsed()

        # Find the Signon class to use. This allows customizing behaviour by
        # subclassing Signon
        if (
            signon_class_path := getattr(settings, "SIGNON_CLASS", None)
        ) is None:
            self.signon_class = Signon
        else:
            self.signon_class = import_string(signon_class_path)
            if not issubclass(self.signon_class, Signon):
                raise ImproperlyConfigured(
                    f"{signon_class_path} is not a subclass of Signon"
                )

        self.get_response = get_response

    def __call__(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponse:
        """Middleware API entry point."""
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "The signon middleware requires the authentication middleware"
                " to be installed.  Edit your MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the SignonMiddleware class."
            )

        # Add request.signon
        cast(RequestSignonProtocol, request).signon = self.signon_class(request)

        return self.get_response(request)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the debdiff workflow."""

from collections.abc import Iterable
from typing import Any, ClassVar, override
from unittest.mock import call, patch

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, Collection, WorkRequest, Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections.lookup import lookup_single
from debusine.server.workflows import DebDiffWorkflow, workflow_utils
from debusine.server.workflows.debdiff import Architecture
from debusine.server.workflows.models import (
    BaseWorkflowData,
    DebDiffWorkflowData,
    DebDiffWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.tasks.models import (
    BaseDynamicTaskData,
    DebDiffFlags,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    LookupSingle,
    OutputData,
    OutputDataError,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class DebDiffWorkflowTests(WorkflowTestCase[DebDiffWorkflow]):
    """Unit tests for :py:class:`DebDiffWorkflow`."""

    origin_source_artifact: ClassVar[Artifact]
    origin_binary_artifacts: ClassVar[dict[Architecture, Artifact]]

    new_source_artifact: ClassVar[Artifact]
    new_binary_artifacts: ClassVar[dict[Architecture, Artifact]]

    debian_collection: ClassVar[Collection]

    workspace: ClassVar[Workspace]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.workspace = cls.playground.get_default_workspace()

        cls.debian_collection = cls.playground.create_collection(
            name="bookworm",
            category=CollectionCategory.SUITE,
        )

        cls.origin_source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0.0"
        )
        cls.new_source_artifact = cls.playground.create_source_artifact(
            name="hello", version="2.0.0", architectures={"all", "amd64"}
        )

        cls.origin_binary_artifacts = {}
        cls.new_binary_artifacts = {}

        for arch in ["amd64", "arm64", "all"]:
            cls.origin_binary_artifacts[Architecture(arch)] = (
                cls.playground.create_minimal_binary_package_artifact(
                    package="hello",
                    srcpkg_name="hello",
                    srcpkg_version="1.0.0",
                    version="1.0.0",
                    architecture=arch,
                )
            )
            cls.new_binary_artifacts[Architecture(arch)] = (
                cls.playground.create_minimal_binary_package_artifact(
                    package="hello",
                    srcpkg_name="hello",
                    srcpkg_version="2.0.0",
                    version="2.0.0",
                    architecture=arch,
                )
            )

    def orchestrate(
        self,
        *,
        task_data: DebDiffWorkflowData,
        sbuild_architectures: list[str],
        source_artifact: Artifact,
        target_distribution: str,
        expect_success: bool = True,
        use_data: bool = False,
    ) -> WorkRequest:
        """Create and orchestrate a DebDiffWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow."""

            @override
            def populate(self) -> None:
                for arch in sbuild_architectures:
                    sbuild, _ = self.work_request_ensure_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            environment=target_distribution,
                            build_architecture="amd64",
                        ),
                        workflow_data=WorkRequestWorkflowData(
                            display_name=f"Build {arch}", step=f"build-{arch}"
                        ),
                    )
                    self.provides_artifact(
                        sbuild,
                        ArtifactCategory.BINARY_PACKAGE,
                        name=f"build-{arch}",
                        data={
                            "binary_package_name": source_artifact.data["name"],
                            "architecture": arch,
                        },
                    )

                debdiff, _ = self.work_request_ensure_child_workflow(
                    task_name="debdiff",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="debdiff",
                        step="debdiff",
                    ),
                )
                self.orchestrate_child(debdiff)

        root = self.playground.create_workflow(task_name="examplepipeline")
        self.assertEqual(
            self.playground.orchestrate_workflow(root, use_data=use_data),
            expect_success,
        )

        return root

    def setup_debdiff_workflow(
        self,
        *,
        add_origin_source_artifact: bool = True,
        add_origin_binary_artifacts_archs: Iterable[str] | None = None,
        override_source_artifact: LookupSingle | None = None,
        add_new_binary_artifacts_archs: Iterable[str] | None = None,
        override_binary_artifacts: list[LookupSingle] | None = None,
        arch_all_build_architecture: str | None = None,
        original: str | None = None,
        extra_flags: list[DebDiffFlags] | None = None,
        expect_success: bool = True,
        use_data: bool = False,
    ) -> None:
        """
        Create and schedule a DebDiff workflow.

        :param add_origin_source_artifact: whether to add
          self.add_origin_source_artifact in the "original" collection
        :param add_origin_binary_artifacts_archs: add each architecture from
          self.origin_binary_artifacts into the "original" collection.
          If None: adds all of them
        :param override_source_artifact: use this source_artifact. If None:
          use default's self.new_source_artifact
        :param add_new_binary_artifacts_archs: add each architecture from
          self.new_binary_artifacts in the DebDiff's "task_data"
          If None: adds all of them
        :param override_binary_artifacts: if not None: for binary_artifacts
          use this LookupSingle (ignoring add_new_binary_artifacts_archs)
        :param arch_all_build_architecture: if None: does nothing. Else add
           "arch_all_build_architecture" in the "task_data"
        :param original: collection to use as "original". If it's None, use
          self.debian_collection
        :param extra_flags: if not None add "extra_flags" in the task data
        :param expect_success: whether the orchestrator is expected to
          succeed
        """
        if add_origin_source_artifact:
            self.debian_collection.manager.add_artifact(
                self.origin_source_artifact,
                user=self.playground.get_default_user(),
                variables={"component": "main", "section": "devel"},
            )

        if add_origin_binary_artifacts_archs is None:
            add_origin_binary_artifacts_archs = (
                self.origin_binary_artifacts.keys()
            )

        if add_new_binary_artifacts_archs is None:
            add_new_binary_artifacts_archs = self.new_binary_artifacts.keys()

        for arch in add_origin_binary_artifacts_archs:
            self.debian_collection.manager.add_artifact(
                self.origin_binary_artifacts[Architecture(arch)],
                user=self.playground.get_default_user(),
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )

        binary_artifacts: list[LookupSingle]
        if override_binary_artifacts is None:
            # Use the default binary artifacts for all the architectures
            binary_artifacts = [
                self.new_binary_artifacts[Architecture(arch)].id
                for arch in add_new_binary_artifacts_archs
            ]

        else:
            # Use the override_binary_artifacts.
            # Calculate the relevant archs based on the passed artifacts
            binary_artifacts = override_binary_artifacts

        sbuild_archs = self.archs_for_binaries(binary_artifacts)

        task_data: dict[str, Any] = {
            "source_artifact": (
                override_source_artifact
                if override_source_artifact is not None
                else self.new_source_artifact.id
            ),
            "binary_artifacts": binary_artifacts,
            "original": original or str(self.debian_collection),
            "vendor": "debian",
            "codename": "bookworm",
        }

        if arch_all_build_architecture is not None:
            task_data["arch_all_build_architecture"] = (
                arch_all_build_architecture
            )

        if extra_flags is not None:
            task_data["extra_flags"] = extra_flags

        example_pipeline = self.orchestrate(
            task_data=DebDiffWorkflowData(**task_data),
            sbuild_architectures=list(sbuild_archs),
            source_artifact=self.new_source_artifact,
            target_distribution="debian:bookworm",
            expect_success=expect_success,
            use_data=use_data,
        )
        work_request = example_pipeline.children.get(
            task_name="debdiff", task_type=TaskTypes.WORKFLOW
        )
        self.task_set_current(self.get_workflow(work_request))

    def archs_for_binaries(self, binaries: list[LookupSingle]) -> set[str]:
        archs = set()
        for lookup in binaries:
            archs |= workflow_utils.lookup_result_architectures(
                lookup_single(
                    lookup,
                    self.workspace,
                    pc=PermissionContext.for_ui_user(
                        self.playground.get_default_user()
                    ),
                    expect_type=LookupChildType.ARTIFACT,
                )
            )
        return archs

    @classmethod
    def new_binary_artifact_lookup(cls, arch: str) -> int:
        return cls.new_binary_artifacts[Architecture(arch)].id

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        # We'll deliberately add a binary package newer than the one we're
        # interested in, so allow this for test purposes
        self.debian_collection.data["may_rewind_versions"] = True
        self.debian_collection.save()

        # Add a binary package with older version: it will be discarded
        old_package_1 = self.playground.create_minimal_binary_package_artifact(
            package="libhello0",
            srcpkg_name="hello",
            srcpkg_version="0.1.0",
            version="0.1.0",
            architecture="amd64",
        )
        self.debian_collection.manager.add_artifact(
            old_package_1,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        old_package_2 = self.playground.create_minimal_binary_package_artifact(
            package="libhello0",
            srcpkg_name="hello",
            srcpkg_version="0.2.0",
            version="0.2.0",
            architecture="amd64",
        )
        self.debian_collection.manager.add_artifact(
            old_package_2,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        # Add another binary package with same version but new architecture
        new_architecture = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello",
                srcpkg_name="hello",
                srcpkg_version="1.0.0",
                version="1.0.0",
                architecture="riscv64",
            )
        )
        self.debian_collection.manager.add_artifact(
            new_architecture,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        # Add a binary package that is ignored because its srcpkg_version
        # is too new (when known, it's not using the most recent binary package
        # but one with a matching srcpkg_version)
        too_new = self.playground.create_minimal_binary_package_artifact(
            package="hello",
            srcpkg_name="hello",
            srcpkg_version="3.0.0",
            version="3.0.0",
            architecture="amd64",
        )
        self.debian_collection.manager.add_artifact(
            too_new,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )

        # Create workflow
        with patch.object(
            DebDiffWorkflow,
            "requires_artifact",
            wraps=DebDiffWorkflow.requires_artifact,
        ) as requires_artifact:
            self.setup_debdiff_workflow(
                arch_all_build_architecture="arm64",
                extra_flags=[DebDiffFlags.IGNORE_SPACE],
            )

        debdiff_work_requests = self.work_request.children.all()
        bookworm = self.debian_collection
        bookworm_lookup = f"{bookworm.name}@{bookworm.category}"

        # Check work request to compare the source packages
        source_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-source"
        )
        self.assertEqual(
            source_work_request.workflow_data.display_name,
            "DebDiff for source package",
        )

        self.assertEqual(
            source_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "arm64",
                "extra_flags": [DebDiffFlags.IGNORE_SPACE],
                "input": {
                    "source_artifacts": [
                        f"{bookworm_lookup}/name:hello_1.0.0",
                        self.new_source_artifact.id,
                    ]
                },
            },
        )

        self.assertIn(
            call(
                source_work_request,
                f"{bookworm_lookup}/name:hello_1.0.0",
            ),
            requires_artifact.call_args_list,
        )
        self.assertIn(
            call(source_work_request, self.new_source_artifact.id),
            requires_artifact.call_args_list,
        )

        # Check work requests to compare the binary packages
        self.assertEqual(
            debdiff_work_requests.filter(
                workflow_data_json__step__startswith="debdiff-binaries"
            ).count(),
            2,
        )

        amd64_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-amd64"
        )

        self.assertEqual(
            amd64_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "amd64",
                "extra_flags": [DebDiffFlags.IGNORE_SPACE],
                "input": {
                    "binary_artifacts": [
                        [
                            f"{bookworm_lookup}/name:hello_1.0.0_all",
                            f"{bookworm_lookup}/name:hello_1.0.0_amd64",
                        ],
                        sorted(
                            [
                                self.new_binary_artifact_lookup("all"),
                                self.new_binary_artifact_lookup("amd64"),
                            ]
                        ),
                    ]
                },
            },
        )

        self.assertIn(
            call(
                amd64_work_request,
                LookupMultiple(
                    (
                        f"{bookworm_lookup}/name:hello_1.0.0_all",
                        f"{bookworm_lookup}/name:hello_1.0.0_amd64",
                    )
                ),
            ),
            requires_artifact.call_args_list,
        )
        self.assertIn(
            call(
                amd64_work_request,
                LookupMultiple.model_validate(
                    sorted(
                        [
                            self.new_binary_artifact_lookup("amd64"),
                            self.new_binary_artifact_lookup("all"),
                        ]
                    )
                ),
            ),
            requires_artifact.call_args_list,
        )

    @preserve_db_task_registry()
    def test_populate_binary_artifacts_upload(self) -> None:
        upload_artifacts = self.playground.create_upload_artifacts(
            binaries=[("hello", "amd64"), ("bye", "amd64")],
        )

        self.setup_debdiff_workflow(
            override_binary_artifacts=[upload_artifacts.upload.id]
        )

        debdiff_work_requests = self.work_request.children.all()

        work_request_source = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-source"
        )
        self.assertEqual(
            work_request_source.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bookworm",
                "extra_flags": [],
                "input": {
                    "source_artifacts": [
                        "bookworm@debian:suite/name:hello_1.0.0",
                        self.new_source_artifact.id,
                    ]
                },
            },
        )

        work_request_binary = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-amd64"
        )

        self.assertEqual(
            work_request_binary.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bookworm",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        [
                            "bookworm@debian:suite/name:hello_1.0.0_all",
                            "bookworm@debian:suite/name:hello_1.0.0_amd64",
                        ],
                        sorted(
                            [
                                artifact.id
                                for artifact in upload_artifacts.binaries
                            ]
                        ),
                    ]
                },
            },
        )

    @preserve_db_task_registry()
    def test_get_new_packages_by_name_arch_invalid_category(self) -> None:
        # Valid artifact categories are tested as part of other tests, such as
        # test_populate_binary_artifacts_upload(). The invalid category
        # path should never be reached in production
        source_artifact = self.playground.create_source_artifact()
        self.setup_debdiff_workflow()

        # Make workflow.data.binary_artifacts contain an artifact of an
        # unexpected category
        self.task.data.binary_artifacts = LookupMultiple((source_artifact.id,))
        with self.assertRaisesRegex(
            ValueError, "Unexpected category 'debian:source-package'"
        ):
            self.task._get_new_packages_by_name_arch()

    @preserve_db_task_registry()
    def test_populate_new_binary_artifacts_only_all_arch(self) -> None:
        self.setup_debdiff_workflow(
            add_new_binary_artifacts_archs=["all"],
            arch_all_build_architecture="arm64",
        )

        debdiff_work_requests = self.work_request.children.all()

        # Check work requests to compare the binary packages
        self.assertEqual(
            debdiff_work_requests.filter(
                workflow_data_json__step__startswith="debdiff-binaries"
            ).count(),
            1,
        )

        all_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-all"
        )

        bookworm = self.debian_collection
        bookworm_lookup = f"{bookworm.name}@{bookworm.category}"

        self.assertEqual(
            all_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "arm64",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        [
                            f"{bookworm_lookup}/name:hello_1.0.0_all",
                        ],
                        [
                            self.new_binary_artifact_lookup("all"),
                        ],
                    ]
                },
            },
        )

    @preserve_db_task_registry()
    def test_populate_origin_binary_artifacts_no_all_arch(self) -> None:
        self.setup_debdiff_workflow(
            add_origin_binary_artifacts_archs=["amd64", "arm64"]
        )

        debdiff_work_requests = self.work_request.children.all()

        # Check work requests to compare the binary packages
        self.assertEqual(
            debdiff_work_requests.filter(
                workflow_data_json__step__startswith="debdiff-binaries"
            ).count(),
            2,
        )

        amd64_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-amd64"
        )

        bookworm = self.debian_collection
        bookworm_lookup = f"{bookworm.name}@{bookworm.category}"

        self.assertEqual(
            amd64_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "amd64",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        [
                            f"{bookworm_lookup}/name:hello_1.0.0_amd64",
                        ],
                        sorted(
                            [
                                self.new_binary_artifact_lookup("amd64"),
                                self.new_binary_artifact_lookup("all"),
                            ]
                        ),
                    ]
                },
            },
        )

    @preserve_db_task_registry()
    def test_populate_new_package(self) -> None:
        """Run populate with a package not found in original collection."""
        self.setup_debdiff_workflow(
            add_origin_source_artifact=False,
            add_origin_binary_artifacts_archs=[],
        )

        # Source package neither binary packages could be found in the
        # original collection. It's a new package, and there isn't any
        # DebDiff task created.
        self.assertEqual(self.work_request.children.count(), 0)

    @preserve_db_task_registry()
    def test_populate_package_without_source_in_original(self) -> None:
        # Add binary packages that have the same name but are older
        # (to exercise returning only the newest)
        for version in ("0.0.1", "0.1.0"):
            self.debian_collection.manager.add_artifact(
                self.playground.create_minimal_binary_package_artifact(
                    package="hello",
                    srcpkg_name="hello",
                    srcpkg_version=version,
                    version=version,
                    architecture="amd64",
                ),
                user=self.playground.get_default_user(),
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )

        self.setup_debdiff_workflow(
            add_origin_source_artifact=False,
        )

        debdiff_work_requests = self.work_request.children.all()

        # Check work requests to compare the binary packages
        self.assertEqual(
            debdiff_work_requests.filter(
                workflow_data_json__step__startswith="debdiff-binaries"
            ).count(),
            2,
        )

        # The version 0.1.0 is not used, only 1.0.0 versions
        for debdiff_work_request in debdiff_work_requests:
            for binary_artifact in debdiff_work_request.task_data["input"][
                "binary_artifacts"
            ][0]:
                self.assertIn("1.0.0", binary_artifact)

    @preserve_db_task_registry()
    def test_populate_source_package_upload_artifact(self) -> None:
        upload_artifact = self.playground.create_upload_artifacts(
            src_name="hello",
            binary=False,
        )

        self.setup_debdiff_workflow(
            override_source_artifact=f"{upload_artifact.upload.id}@artifacts"
        )

        source_work_request = self.work_request.children.get(
            workflow_data_json__step="debdiff-source"
        )

        new_source_artifact = source_work_request.task_data["input"][
            "source_artifacts"
        ][1]

        self.assertEqual(
            new_source_artifact, f"{upload_artifact.source.id}@artifacts"
        )

    @preserve_db_task_registry()
    def test_populate_one_debdiff_per_concrete_arch_includes_all_binaries(
        self,
    ) -> None:
        self.setup_debdiff_workflow()

        debdiff_work_requests = self.work_request.children.all()

        expected_archs = {"amd64", "arm64"}

        binary_wrs = debdiff_work_requests.filter(
            workflow_data_json__step__startswith="debdiff-binaries"
        )

        # One per expected concrete arch (without "all")
        steps = binary_wrs.values_list("workflow_data_json__step", flat=True)
        self.assertCountEqual(
            steps,
            {f"debdiff-binaries-{arch}" for arch in expected_archs},
        )

        bookworm = self.debian_collection
        bookworm_lookup = f"{bookworm.name}@{bookworm.category}"

        for arch in expected_archs:
            wr = binary_wrs.get(
                workflow_data_json__step=f"debdiff-binaries-{arch}"
            )
            originals, news = wr.task_data["input"]["binary_artifacts"]

            # Originals: "all" + specific architecture packages
            self.assertEqual(
                set(originals),
                {
                    f"{bookworm_lookup}/name:hello_1.0.0_all",
                    f"{bookworm_lookup}/name:hello_1.0.0_{arch}",
                },
            )

            # news: "all" + specific architecture packages
            self.assertEqual(
                set(news),
                {
                    self.new_binary_artifact_lookup("all"),
                    self.new_binary_artifact_lookup(arch),
                },
            )

    @preserve_db_task_registry()
    def test_populate_binary_for_all_does_not_include_arch_specific_originals(
        self,
    ) -> None:
        self.setup_debdiff_workflow(
            add_new_binary_artifacts_archs=["all"],
        )

        debdiff_work_requests = self.work_request.children.all()
        all_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-all"
        )

        originals, _ = all_work_request.task_data["input"]["binary_artifacts"]
        debian_collection_lookup = (
            f"{self.debian_collection.name}@{self.debian_collection.category}"
        )

        self.assertEqual(
            originals, [f"{debian_collection_lookup}/name:hello_1.0.0_all"]
        )

    @preserve_db_task_registry()
    def test_populate_origin_architecture_from_any_to_all(self) -> None:
        """Binary debdiff tasks created if binaries change from any to all."""
        self.setup_debdiff_workflow(
            add_origin_binary_artifacts_archs=["amd64", "arm64"],
            add_new_binary_artifacts_archs=["all"],
        )

        debdiff_work_requests = self.work_request.children.all()

        binary_wrs = debdiff_work_requests.filter(
            workflow_data_json__step__startswith="debdiff-binaries"
        )

        expected_archs = {"amd64", "arm64"}

        steps = binary_wrs.values_list("workflow_data_json__step", flat=True)

        # Correct archs
        self.assertCountEqual(
            steps,
            {f"debdiff-binaries-{arch}" for arch in expected_archs},
        )

        bookworm = self.debian_collection
        bookworm_lookup = f"{bookworm.name}@{bookworm.category}"

        # Correct task_data per architecture
        amd64_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-amd64"
        )

        self.assertEqual(
            amd64_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "amd64",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        [f"{bookworm_lookup}/name:hello_1.0.0_amd64"],
                        [self.new_binary_artifact_lookup("all")],
                    ]
                },
            },
        )

        arm64_work_request = debdiff_work_requests.get(
            workflow_data_json__step="debdiff-binaries-arm64"
        )

        self.assertEqual(
            arm64_work_request.task_data,
            {
                "environment": "debian/match:codename=bookworm",
                "build_architecture": "arm64",
                "extra_flags": [],
                "input": {
                    "binary_artifacts": [
                        [f"{bookworm_lookup}/name:hello_1.0.0_arm64"],
                        [self.new_binary_artifact_lookup("all")],
                    ]
                },
            },
        )

    @preserve_db_task_registry()
    def test_compute_scheduler_tags(self) -> None:
        self.setup_debdiff_workflow()
        self.assertComputesSchedulerTags()

    @preserve_db_task_registry()
    def test_compute_dynamic_data_from_source(self) -> None:
        self.setup_debdiff_workflow()
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffWorkflowDynamicData(
                subject="hello",
                parameter_summary="bookworm: 1.0.0 → 2.0.0",
                original_id=self.origin_source_artifact.id,
                source_artifact_id=self.new_source_artifact.id,
                binary_artifacts_ids=[
                    artifact.id
                    for artifact in self.new_binary_artifacts.values()
                ],
            ),
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts(self) -> None:
        self.setup_debdiff_workflow()
        self.task.dynamic_data = self.compute_dynamic_data(self.task)
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.new_source_artifact.id,
                ),
                InputArtifactSingle(
                    lookup=self.task.data.original,
                    label="original",
                    artifact_id=self.origin_source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[
                        artifact.id
                        for artifact in self.new_binary_artifacts.values()
                    ],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_compute_dynamic_data_from_binary(self) -> None:
        # Add a binary package with older version: it will be discarded
        old_package = self.playground.create_minimal_binary_package_artifact(
            package="libhello0",
            srcpkg_name="hello",
            srcpkg_version="0.1.0",
            version="0.1.0",
            architecture="amd64",
        )
        self.debian_collection.manager.add_artifact(
            old_package,
            user=self.playground.get_default_user(),
            variables={
                "component": "main",
                "section": "devel",
                "priority": "optional",
            },
        )
        self.setup_debdiff_workflow(add_origin_source_artifact=False)

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffWorkflowDynamicData(
                subject="hello",
                parameter_summary="bookworm: 1.0.0 → 2.0.0",
                original_id=None,
                source_artifact_id=self.new_source_artifact.id,
                binary_artifacts_ids=[
                    artifact.id
                    for artifact in self.new_binary_artifacts.values()
                ],
            ),
        )

    @preserve_db_task_registry()
    def test_compute_dynamic_data_package_not_in_origin_collection(
        self,
    ) -> None:
        self.setup_debdiff_workflow(
            add_origin_source_artifact=False,
            add_origin_binary_artifacts_archs=[],
        )
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            DebDiffWorkflowDynamicData(
                subject="hello",
                parameter_summary="bookworm: missing → 2.0.0",
                original_id=None,
                source_artifact_id=self.new_source_artifact.id,
                binary_artifacts_ids=[
                    artifact.id
                    for artifact in self.new_binary_artifacts.values()
                ],
            ),
        )

    def test_package_arch_pairs_artifact_debian_binary_package(self) -> None:
        temp_dir = self.create_temporary_directory()

        artifact = self.playground.create_artifact_from_local(
            self.playground.create_binary_package(
                temp_dir, name="hello", architecture="arm64"
            )
        )

        lookup_result = lookup_single(
            f"{artifact.id}@artifacts",
            workspace=self.workspace,
            pc=PermissionContext.for_ui_user(
                self.playground.get_default_user()
            ),
        )

        self.assertEqual(
            DebDiffWorkflow._package_arch_pairs(lookup_result),
            [("hello", "arm64")],
        )

    def test_package_arch_pairs_artifact_raise_value_error(self) -> None:
        self.playground.create_bare_data_item(
            self.debian_collection,
            "build-amd64",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.TEST,
            },
        )

        lookup_result = lookup_single(
            f"{self.debian_collection}/name:build-amd64",
            self.workspace,
            pc=PermissionContext.for_ui_user(
                self.playground.get_default_user()
            ),
        )

        with self.assertRaisesRegex(
            ValueError,
            r"^DebDiffWorkflow requires binary artifacts \(not promises\)$",
        ):
            DebDiffWorkflow._package_arch_pairs(lookup_result)

    def test_package_arch_pairs_artifact_unsupported_category(self) -> None:
        artifact = self.playground.create_source_artifact()

        lookup_result = lookup_single(
            f"{artifact.id}@artifacts",
            workspace=self.workspace,
            pc=PermissionContext.for_ui_user(
                self.playground.get_default_user()
            ),
        )

        with self.assertRaisesRegex(
            ValueError, rf"Unexpected category {artifact.category}"
        ):
            DebDiffWorkflow._package_arch_pairs(lookup_result)

    def test_package_arch_pairs_artifact_upload(self) -> None:
        # Introspects the relations
        upload_artifacts = self.playground.create_upload_artifacts(
            binaries=[("bye", "all"), ("hello", "all")],
        )

        lookup_result = lookup_single(
            f"{upload_artifacts.upload.id}@artifacts",
            workspace=self.workspace,
            pc=PermissionContext.for_ui_user(
                self.playground.get_default_user()
            ),
        )
        artifact_amd64 = upload_artifacts.binaries[0]
        artifact_amd64.data["deb_fields"]["Architecture"] = "amd64"
        artifact_amd64.save()

        artifact_1 = upload_artifacts.binaries[1]

        self.assertEqual(
            DebDiffWorkflow._package_arch_pairs(lookup_result),
            sorted(
                [
                    (artifact_amd64.data["deb_fields"]["Package"], "amd64"),
                    (artifact_1.data["deb_fields"]["Package"], "all"),
                ]
            ),
        )

    @preserve_db_task_registry()
    def test_validate_input(self) -> None:
        # validate_input() is called and does not raise
        self.setup_debdiff_workflow()

    @preserve_db_task_registry()
    def test_validate_input_invalid_original(self) -> None:
        self.setup_debdiff_workflow(
            original="non-existing", expect_success=False, use_data=True
        )
        # TODO: This should arguably be logged on the debdiff workflow
        # itself, not on its parent.  At the moment,
        # Workflow.orchestrate_child calls mark_pending which fails, and the
        # exception is handled by the parent.
        assert self.work_request.parent is not None
        self.assertEqual(
            self.work_request.parent.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=(
                            "Orchestrator failed: 'non-existing' does not "
                            "specify a category and the context does not "
                            "supply a default"
                        ),
                        code="orchestrator-failed",
                    )
                ]
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the SignRepositoryIndex task."""

import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, override
from unittest import mock

import gpg
from django.test import TestCase as DjangoTestCase
from django.test import TransactionTestCase, override_settings
from nacl.public import PrivateKey

import debusine.worker.tags as wtags
from debusine.artifacts import (
    LocalArtifact,
    RepositoryIndex,
    WorkRequestDebugLogs,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianRepositoryIndex,
    EmptyArtifactData,
    WorkRequestResults,
)
from debusine.assets.models import AssetCategory
from debusine.client.models import (
    ArtifactResponse,
    AssetPermissionCheckResponse,
    RemoteArtifact,
)
from debusine.signing.db.models import Key
from debusine.signing.gnupg import gpg_ephemeral_context
from debusine.signing.models import SigningMode
from debusine.signing.tasks import SignRepositoryIndex
from debusine.signing.tasks.models import SignRepositoryIndexDynamicData
from debusine.signing.tasks.tests.helpers import SigningTaskTestCase
from debusine.tasks.models import InputCollectionSingle
from debusine.test import TestCase
from debusine.test.test_utils import create_remote_artifact


class SignRepositoryIndexTests(
    SigningTaskTestCase[SignRepositoryIndex],
    TestCase,
    DjangoTestCase,
):
    """Unit tests for the :py:class:`SignRepositoryIndex` task."""

    SAMPLE_TASK_DATA = {
        "suite_collection": f"sid@{CollectionCategory.SUITE}",
        "unsigned": 2,
        "mode": "clear",
        "signed_name": "InRelease",
    }

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "unsigned",
            id=2,
            category=ArtifactCategory.REPOSITORY_INDEX,
            data=DebianRepositoryIndex(path="dists/sid/Release"),
        )
        self.resolver.field_values["suite_archive"] = [
            InputCollectionSingle(
                collection_id=2,
                scope_name="scope",
                workspace_name="workspace",
                category=CollectionCategory.SUITE,
                name="sid",
                data={},
                lookup="TODO",
                label="test",
            ),
            InputCollectionSingle(
                collection_id=1,
                scope_name="scope",
                workspace_name="workspace",
                category=CollectionCategory.ARCHIVE,
                name="_",
                data={"signing_keys": ["KEY_ARCHIVE"]},
                lookup="TODO",
                label="test",
            ),
        ]

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "signing:signrepositoryindex:version:1"],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_get_subject(self) -> None:
        self.assertIsNone(self.get_subject())

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_compute_dynamic_data_signing_keys_from_suite(self) -> None:
        self.resolver.field_values["suite_archive"] = [
            InputCollectionSingle(
                collection_id=2,
                scope_name="scope",
                workspace_name="workspace",
                category=CollectionCategory.SUITE,
                name="sid",
                data={"signing_keys": ["KEY_SUITE"]},
                lookup="TODO",
                label="test",
            ),
            InputCollectionSingle(
                collection_id=1,
                scope_name="scope",
                workspace_name="workspace",
                category=CollectionCategory.ARCHIVE,
                name="_",
                data={"signing_keys": ["KEY_ARCHIVE"]},
                lookup="TODO",
                label="test",
            ),
        ]

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SignRepositoryIndexDynamicData(
                signing_keys=["KEY_SUITE"], unsigned_id=2
            ),
        )

    def test_compute_dynamic_data_signing_keys_from_archive(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SignRepositoryIndexDynamicData(
                signing_keys=["KEY_ARCHIVE"], unsigned_id=2
            ),
        )

    def test_compute_dynamic_data_no_archive_or_signing_keys(self) -> None:
        self.resolver.field_values["suite_archive"] = [
            InputCollectionSingle(
                collection_id=1,
                scope_name="scope",
                workspace_name="workspace",
                category=CollectionCategory.SUITE,
                name="sid",
                data={},
                lookup="TODO",
                label="test",
            ),
            None,
        ]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            SignRepositoryIndexDynamicData(signing_keys=[], unsigned_id=2),
        )

    def test_get_input_artifacts_ids(self) -> None:
        self.assertEqual(self.task.get_input_artifacts_ids(), [])
        self.task.dynamic_data = SignRepositoryIndexDynamicData(
            signing_keys=[], unsigned_id=1
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1])

    def test_fetch_input_no_signing_keys(self) -> None:
        directory = self.create_temporary_directory()
        (release := directory / "Release").write_text("A Release file\n")
        unsigned = RepositoryIndex.create(file=release, path="Release")
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={"signing_keys": [], "unsigned_id": unsigned_id},
        )
        self.task_set_current(task)
        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = unsigned_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"No signing keys are configured for "
            f"sid@{CollectionCategory.SUITE}.\n",
        )
        debusine_mock.download_artifact.assert_not_called()

    def test_fetch_input_permission_denied(self) -> None:
        fingerprint = "0" * 64
        directory = self.create_temporary_directory()
        (release := directory / "Release").write_text("A Release file\n")
        unsigned = RepositoryIndex.create(file=release, path="Release")
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={
                "signing_keys": [fingerprint],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)
        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = unsigned_response
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=False,
                username="testuser",
                user_id=123,
                resource={},
            )
        )
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"User testuser does not have permission to sign "
            f"{unsigned_id} with openpgp-repository:{fingerprint}.\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            unsigned_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_mismatching_permission_check_responses(self) -> None:
        keys = [
            Key.objects.create(
                purpose=Key.Purpose.OPENPGP_REPOSITORY,
                fingerprint=fingerprint,
                private_key={},
                public_key=b"",
            )
            for fingerprint in ("0" * 64, "1" * 64)
        ]
        directory = self.create_temporary_directory()
        (release := directory / "Release").write_text("A Release file\n")
        unsigned = RepositoryIndex.create(file=release, path="Release")
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={
                "signing_keys": [key.fingerprint for key in keys],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)
        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = unsigned_response
        asset_permission_checks = [
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={"path": "Release"},
            ),
            AssetPermissionCheckResponse(
                has_permission=True,
                username="somehow-different",
                user_id=123,
                resource={"path": "Release"},
            ),
        ]
        debusine_mock.asset_permission_check.side_effect = (
            asset_permission_checks
        )
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"Mismatching permission-check responses for different signing "
            f"keys: "
            f"{asset_permission_checks[1]} != {asset_permission_checks[0]}\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            unsigned_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_fetch_input_missing_key(self) -> None:
        """fetch_input requires the signing key to exist in the database."""
        fingerprint = "0" * 64
        directory = self.create_temporary_directory()
        (release := directory / "Release").write_text("A Release file\n")
        unsigned = RepositoryIndex.create(file=release, path="Release")
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={
                "signing_keys": [fingerprint],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)
        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.return_value = unsigned_response
        destination = self.create_temporary_directory()

        self.assertFalse(task.fetch_input(destination))

        self.assertEqual(
            (task.log_directory / "fetch_input.log").read_text(),
            f"Signing key openpgp-repository:{fingerprint} does not exist\n",
        )
        debusine_mock.download_artifact.assert_called_once_with(
            unsigned_id,
            destination,
            tarball=mock.ANY,
            workspace=task.workspace_name,
        )

    def test_run_error(self) -> None:
        """run() raises an exception if signing fails."""
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_REPOSITORY,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        self.task._file_path = "Release"
        self.task._keys = [key]
        self.task._permission_check = AssetPermissionCheckResponse(
            has_permission=True,
            username="testuser",
            user_id=123,
            resource={"path": "Release"},
        )
        self.task.prepare_to_run(download_directory, execute_directory)

        with (
            mock.patch(
                "debusine.signing.tasks.sign_repository_index.sign",
                side_effect=ValueError("Boom"),
            ),
            self.assertRaisesRegex(ValueError, "Boom"),
        ):
            self.task.run(execute_directory)

        self.assertIsNone(self.task._signed_path)

    def test_run_success(self) -> None:
        """run() records a successful result if signing succeeds."""
        key = Key.objects.create(
            purpose=Key.Purpose.OPENPGP_REPOSITORY,
            fingerprint="0" * 64,
            private_key={},
            public_key=b"",
        )
        download_directory = self.create_temporary_directory()
        execute_directory = self.create_temporary_directory()
        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": 2,
                "mode": "clear",
                "signed_name": "InRelease",
            }
        )
        self.task_set_current(task)
        task._file_path = "Release"
        task._keys = [key]
        task._permission_check = AssetPermissionCheckResponse(
            has_permission=True,
            username="testuser",
            user_id=123,
            resource={"path": "Release"},
        )
        task.prepare_to_run(download_directory, execute_directory)

        with mock.patch(
            "debusine.signing.tasks.sign_repository_index.sign"
        ) as mock_sign:
            self.assertEqual(
                task.run(execute_directory), WorkRequestResults.SUCCESS
            )

        self.assertEqual(
            task._signed_path, execute_directory / "output" / "InRelease"
        )
        mock_sign.assert_called_once_with(
            [key],
            execute_directory / "input" / "Release",
            execute_directory / "output" / "InRelease",
            SigningMode.CLEAR,
            task.work_request_id,
            username="testuser",
            user_id=123,
            resource={"path": "Release"},
            log_file=mock.ANY,
        )


class SignRepositoryIndexIntegrationTests(
    SigningTaskTestCase[SignRepositoryIndex], TestCase, TransactionTestCase
):
    """Integration tests for the :py:class:`SignRepositoryIndex` task."""

    def generate_signing_key(self, service_private_key: PrivateKey) -> Key:
        """Generate a signing key."""
        with (
            override_settings(
                DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
            ),
            open(os.devnull, "wb") as log_file,
        ):
            key = Key.objects.generate(
                Key.Purpose.OPENPGP_REPOSITORY,
                "An OpenPGP repository key",
                1,
                log_file,
            )
        return key

    def make_release(self) -> tuple[Path, RepositoryIndex]:
        """Prepare a minimal repository index: return (path, artifact)."""
        temp_path = self.create_temporary_directory()
        (release_path := temp_path / "Release").write_text("A Release file\n")
        index = RepositoryIndex.create(file=release_path, path="Release")
        return release_path, index

    def test_execute_detached_success(self) -> None:
        """Integration test: detached-sign using real OpenPGP keys."""
        service_private_key = PrivateKey.generate()
        keys = [
            self.generate_signing_key(service_private_key) for _ in range(2)
        ]
        release_path, unsigned = self.make_release()
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)
        uploaded_paths: list[Path] = []
        uploaded_signed: RepositoryIndex | None = None
        output_path = self.create_temporary_directory()
        uploaded_artifact_ids = {
            ArtifactCategory.REPOSITORY_INDEX: 3,
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: 4,
        }

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == unsigned_id
            shutil.copy(release_path, destination)
            return unsigned_response

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            nonlocal uploaded_signed
            for path in local_artifact.files.values():
                uploaded_paths.append(path)
                shutil.copy(path, output_path)
            if local_artifact.category == ArtifactCategory.REPOSITORY_INDEX:
                [(signed_path_in_suite, signed_path)] = (
                    local_artifact.files.items()
                )
                uploaded_signed = RepositoryIndex.create(
                    file=signed_path,
                    path=signed_path_in_suite,
                    content_type="application/pgp-signature; charset=utf-8",
                )
            return create_remote_artifact(
                id=uploaded_artifact_ids[local_artifact.category],
                workspace="System",
            )

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "detached",
                "signed_name": "Release.gpg",
            },
            dynamic_task_data={
                "signing_keys": [key.fingerprint for key in keys],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={"path": "Release"},
            )
        )

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)

        self.assertEqual(
            debusine_mock.asset_permission_check.call_count, len(keys)
        )
        debusine_mock.asset_permission_check.assert_has_calls(
            [
                mock.call(
                    asset_category=AssetCategory.SIGNING_KEY,
                    asset_slug=f"openpgp-repository:{key.fingerprint}",
                    permission_name="sign_repository_with",
                    artifact_id=unsigned_id,
                    work_request_id=task.work_request_id,
                    workspace=task.workspace_name,
                )
                for key in keys
            ]
        )

        # The task uploaded an artifact with the expected properties.
        self.assertEqual(len(uploaded_paths), 2)
        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    uploaded_signed,
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    WorkRequestDebugLogs(
                        category=WorkRequestDebugLogs._category,
                        data=EmptyArtifactData(),
                        files={
                            "cmd-output.log": (
                                task.log_directory / "cmd-output.log"
                            )
                        },
                        content_types={
                            "cmd-output.log": "text/plain; charset=utf-8"
                        },
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_any_call(
            uploaded_artifact_ids[ArtifactCategory.REPOSITORY_INDEX],
            unsigned_id,
            "extends",
        )

        # The file was properly signed.
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            for key in keys:
                import_result = ctx.key_import(bytes(key.public_key))
                assert getattr(import_result, "imported", 0) == 1
            _, result = ctx.verify(
                release_path.read_bytes(),
                (output_path / "Release.gpg").read_bytes(),
            )
            self.assertEqual(len(result.signatures), len(keys))
            self.assertCountEqual(
                [signature.fpr for signature in result.signatures],
                [key.fingerprint for key in keys],
            )
            # We configure this in gpg_ephemeral_context, but make sure of
            # it.
            self.assertEqual(
                [signature.hash_algo for signature in result.signatures],
                [gpg.constants.MD_SHA512] * len(keys),
            )

    def test_execute_clear_success(self) -> None:
        """Integration test: clear-sign using real OpenPGP keys."""
        service_private_key = PrivateKey.generate()
        keys = [
            self.generate_signing_key(service_private_key) for _ in range(2)
        ]
        release_path, unsigned = self.make_release()
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)
        uploaded_paths: list[Path] = []
        uploaded_signed: RepositoryIndex | None = None
        output_path = self.create_temporary_directory()
        uploaded_artifact_ids = {
            ArtifactCategory.REPOSITORY_INDEX: 3,
            ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: 4,
        }

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == unsigned_id
            shutil.copy(release_path, destination)
            return unsigned_response

        def upload_artifact(
            local_artifact: LocalArtifact[Any], **kwargs: Any
        ) -> RemoteArtifact:
            nonlocal uploaded_signed
            for path in local_artifact.files.values():
                uploaded_paths.append(path)
                shutil.copy(path, output_path)
            if local_artifact.category == ArtifactCategory.REPOSITORY_INDEX:
                [(signed_path_in_suite, signed_path)] = (
                    local_artifact.files.items()
                )
                uploaded_signed = RepositoryIndex.create(
                    file=signed_path,
                    path=signed_path_in_suite,
                    content_type="text/PGP; charset=utf-8",
                )
            return create_remote_artifact(
                id=uploaded_artifact_ids[local_artifact.category],
                workspace="testing",
            )

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={
                "signing_keys": [key.fingerprint for key in keys],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact
        debusine_mock.upload_artifact.side_effect = upload_artifact
        debusine_mock.asset_permission_check.return_value = (
            AssetPermissionCheckResponse(
                has_permission=True,
                username="testuser",
                user_id=123,
                resource={"path": "Release"},
            )
        )

        with override_settings(
            DEBUSINE_SIGNING_PRIVATE_KEYS=[service_private_key]
        ):
            self.assertEqual(task.execute(), WorkRequestResults.SUCCESS)

        self.assertEqual(
            debusine_mock.asset_permission_check.call_count, len(keys)
        )
        debusine_mock.asset_permission_check.assert_has_calls(
            [
                mock.call(
                    asset_category=AssetCategory.SIGNING_KEY,
                    asset_slug=f"openpgp-repository:{key.fingerprint}",
                    permission_name="sign_repository_with",
                    artifact_id=unsigned_id,
                    work_request_id=task.work_request_id,
                    workspace=task.workspace_name,
                )
                for key in keys
            ]
        )

        # The task uploaded an artifact with the expected properties.
        self.assertEqual(len(uploaded_paths), 2)
        debusine_mock.upload_artifact.assert_has_calls(
            [
                mock.call(
                    uploaded_signed,
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
                mock.call(
                    WorkRequestDebugLogs(
                        category=WorkRequestDebugLogs._category,
                        data=EmptyArtifactData(),
                        files={
                            "cmd-output.log": (
                                task.log_directory / "cmd-output.log"
                            )
                        },
                        content_types={
                            "cmd-output.log": "text/plain; charset=utf-8"
                        },
                    ),
                    workspace=task.workspace_name,
                    work_request=task.work_request_id,
                ),
            ]
        )
        debusine_mock.relation_create.assert_any_call(
            uploaded_artifact_ids[ArtifactCategory.REPOSITORY_INDEX],
            unsigned_id,
            "extends",
        )

        # The file was properly signed.
        with (
            TemporaryDirectory(prefix="debusine-tests-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            for key in keys:
                import_result = ctx.key_import(bytes(key.public_key))
                assert getattr(import_result, "imported", 0) == 1
            data, result = ctx.verify((output_path / "InRelease").read_bytes())
            self.assertEqual(data, b"A Release file\n")
            self.assertEqual(len(result.signatures), len(keys))
            self.assertCountEqual(
                [signature.fpr for signature in result.signatures],
                [key.fingerprint for key in keys],
            )
            # We configure this in gpg_ephemeral_context, but make sure of
            # it.
            self.assertEqual(
                [signature.hash_algo for signature in result.signatures],
                [gpg.constants.MD_SHA512] * len(keys),
            )

    def test_execute_failure(self) -> None:
        """If execution fails, the task does not upload an output artifact."""
        service_private_key = PrivateKey.generate()
        key = self.generate_signing_key(service_private_key)
        release_path, unsigned = self.make_release()
        unsigned_id = 2
        unsigned_response = self.make_artifact_response(unsigned, unsigned_id)

        def download_artifact(
            artifact_id: int, destination: Path, **kwargs: Any
        ) -> ArtifactResponse:
            assert artifact_id == unsigned_id
            shutil.copy(release_path, destination)
            return unsigned_response

        task = SignRepositoryIndex(
            task_data={
                "suite_collection": f"sid@{CollectionCategory.SUITE}",
                "unsigned": unsigned_id,
                "mode": "clear",
                "signed_name": "InRelease",
            },
            dynamic_task_data={
                "signing_keys": [key.fingerprint],
                "unsigned_id": unsigned_id,
            },
        )
        self.task_set_current(task)

        debusine_mock = self.mock_debusine(task)
        debusine_mock.download_artifact.side_effect = download_artifact

        with mock.patch(
            "debusine.signing.tasks.sign_repository_index.sign",
            side_effect=Exception("Signing failed"),
        ):
            self.assertEqual(task.execute(), WorkRequestResults.FAILURE)

        # Only a debug logs artifact was uploaded, not any signing output
        debusine_mock.upload_artifact.assert_called_once_with(
            WorkRequestDebugLogs(
                category=WorkRequestDebugLogs._category,
                data=EmptyArtifactData(),
                files={
                    "stages.log": (task.log_directory / "stages.log"),
                    "cmd-output.log": (task.log_directory / "cmd-output.log"),
                    "execution.log": (task.log_directory / "execution.log"),
                },
                content_types={
                    "stages.log": "text/plain; charset=utf-8",
                    "cmd-output.log": "text/plain; charset=utf-8",
                    "execution.log": "text/plain; charset=utf-8",
                },
            ),
            workspace=task.workspace_name,
            work_request=task.work_request_id,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Celery integration for debusine tasks."""

import logging
from contextlib import ExitStack

from celery import shared_task
from django.db import transaction

from debusine.artifacts.models import TaskTypes
from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.server.tasks import BaseServerTask
from debusine.tasks import TaskConfigError
from debusine.tasks.models import OutputData, OutputDataError

logger = logging.getLogger(__name__)


class WorkRequestNotPending(Exception):
    """We only run pending work requests."""


class ServerTaskRunError(Exception):
    """Running a server task failed."""

    def __init__(self, message: str, code: str) -> None:
        """Construct the exception."""
        self.message = message
        self.code = code


def _run_server_task_or_error(work_request: WorkRequest) -> WorkRequest.Results:
    """Run a server task, raising :py:class:`ServerTaskRunError` on errors."""
    if work_request.task_type != TaskTypes.SERVER:
        raise ServerTaskRunError(
            "Cannot run on a Celery worker", "wrong-task-type"
        )

    task_name = work_request.task_name
    try:
        task = work_request.get_task()
    except ValueError as exc:
        raise ServerTaskRunError(str(exc), "setup-failed") from exc
    except TaskConfigError as exc:
        raise ServerTaskRunError(
            f"Failed to configure: {exc}", "configure-failed"
        ) from exc
    assert isinstance(task, BaseServerTask)

    try:
        with ExitStack() as stack:
            if not task.TASK_MANAGES_TRANSACTIONS:
                stack.enter_context(transaction.atomic())
            result = task.execute_logging_exceptions()
    except Exception as exc:
        raise ServerTaskRunError(
            f"Execution failed: {exc}", "execute-failed"
        ) from exc
    else:
        if task.aborted:
            logger.info("Task: %s has been aborted", task_name)
            # No need to update DB state
            return WorkRequest.Results.FAILURE
        else:
            # Lock the whole workflow and mark the work request as
            # completed.
            with transaction.atomic():
                WorkRequest.objects.filter(
                    id=work_request.id
                ).lock_workflows_for_update()
                work_request.mark_completed(result)
                return result


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
# Unlike some other Celery tasks, this must _not_ be decorated with
# @transaction.atomic, because it needs to do finer-grained transaction
# control.
def run_server_task(work_request_id: int) -> WorkRequest.Results:
    """
    Run a :py:class:`BaseServerTask` via Celery.

    :param work_request_id: The ID of the work request to run.
    :raises WorkRequest.DoesNotExist: if ``work_request_id`` does not exist.
    :raises WorkRequestNotPending: if the given work request's status is not
      "pending".
    :return: SUCCESS if the server task succeeded; ERROR if it failed or raised
      an exception.
    """
    try:
        work_request = WorkRequest.objects.get(pk=work_request_id)
    except WorkRequest.DoesNotExist:
        logger.error("Work request %d does not exist", work_request_id)
        raise

    if work_request.status != WorkRequest.Statuses.PENDING:
        logger.error(
            "Work request %d is in status %s, not pending",
            work_request_id,
            work_request.status,
        )
        raise WorkRequestNotPending

    # We would normally lock the whole workflow as close to the start of the
    # operation as possible, holding that lock until the end of the
    # transaction to minimize the risk of deadlocks when updating multiple
    # work requests in succession.  However, server tasks may take a long
    # time to run, and some of them manage their own transactions.
    #
    # WorkRequest.mark_running has no cascading effects on other work
    # requests, and at this point we're in Django's default autocommit mode,
    # so the worst case at this point is that it fails to obtain a lock on
    # this work request; it cannot cause a deadlock.
    work_request.mark_running()

    context.reset()
    work_request.set_current()

    try:
        return _run_server_task_or_error(work_request)
    except ServerTaskRunError as exc:
        # Lock the whole workflow and record the error.
        with transaction.atomic():
            logger.error(
                "Error running work request %s/%s (%s): %s",
                work_request.task_type,
                work_request.task_name,
                work_request.id,
                exc.message,
            )
            WorkRequest.objects.filter(
                id=work_request.id
            ).lock_workflows_for_update()
            work_request.mark_completed(
                WorkRequest.Results.ERROR,
                output_data=OutputData(
                    errors=[OutputDataError(message=exc.message, code=exc.code)]
                ),
            )
            return WorkRequest.Results.ERROR



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for account (log in, log out)."""

from django.test import TestCase
from django.urls import reverse

from debusine.web.views.tests.utils import ViewTestMixin


class LoginTests(ViewTestMixin, TestCase):
    """Test login functionality and template."""

    def test_view(self) -> None:
        """Generic test of the view."""
        next_path = "/workers"
        response = self.client.get(reverse("login") + f"?next={next_path}")
        tree = self.assertResponseHTML(response)

        # Correct title
        self.assertContains(response, "Log in")

        # Labels for the form
        self.assertContains(response, "Username *")
        self.assertContains(response, "Password *")

        # Has "next" form hidden field
        input_next = tree.xpath(
            ".//form[@id='login-form']//input[@name='next' and @type='hidden']"
        )

        self.assertEqual(len(input_next), 1)

        self.assertEqual(input_next[0].attrib["value"], next_path)

    def test_login_form_errors(self) -> None:
        """If the form has errors: template renders them."""
        response = self.client.post(
            reverse("login"), {"username": "", "password": ""}
        )

        self.assertContains(
            response,
            "Your username and password didn't match. Please try again.",
        )
        self.assertTrue(response.context["form"].errors)

    def test_view_login_without_next(self) -> None:
        """
        Base templates do not include ?next=login path.

        URL for login add is_login_view, base templates use the variable to not
        add ?next={next_path}
        """
        response = self.client.get(reverse("login"))

        next_path = reverse("login")
        self.assertNotContains(response, f"?next={next_path}")


class LogoutTests(ViewTestMixin, TestCase):
    """Tests related to logout functionality."""

    def test_view_login_without_next(self) -> None:
        """
        Base templates do not include ?next=logout path.

        URL for logout add is_logout_view, base templates use the variable to
        not add ?next={next_path}
        """
        response = self.client.post(reverse("logout"))

        next_path = reverse("logout")
        self.assertNotContains(response, f"?next={next_path}")

    def test_view_login_with_next(self) -> None:
        """Base templates do include ?next=workspaces_list path."""
        next_path = reverse("scopes:detail")
        response = self.client.get(next_path)
        self.assertContains(response, f"?next={next_path}")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the models."""

from typing import override

from django.core.exceptions import ValidationError

from debusine.db.models import NotificationChannel
from debusine.test.django import TestCase


class NotificationChannelTests(TestCase):
    """Tests for NotificationChannel class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.email_data = {
            "from": "sender@debusine.example.org",
            "to": ["recipient@example.com"],
        }

    def test_create_email(self) -> None:
        """Create email NotificationChannel."""
        name = "deblts-email"
        method = NotificationChannel.Methods.EMAIL

        notification_channel = NotificationChannel.objects.create(
            name=name, method=method, data=self.email_data
        )

        notification_channel.refresh_from_db()

        self.assertEqual(notification_channel.name, name)
        self.assertEqual(notification_channel.method, method)
        self.assertEqual(notification_channel.data, self.email_data)

    def test_create_email_invalid_data(self) -> None:
        """Create email NotificationChannel: invalid data, failure."""
        with self.assertRaises(ValidationError):
            NotificationChannel.objects.create(
                name="deblts-email",
                method=NotificationChannel.Methods.EMAIL,
                data={"something": "something"},
            )

    def test_name_unique(self) -> None:
        """Field name is unique."""
        name = "deblts-email"
        NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        with self.assertRaises(ValidationError):
            NotificationChannel.objects.create(
                name=name,
                method=NotificationChannel.Methods.EMAIL,
                data=self.email_data,
            )

    def test_str(self) -> None:
        """Assert __str__ return the expected string."""
        name = "deblts-email"
        notification_channel = NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        self.assertEqual(str(notification_channel), name)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tar-related classes: generate tar.gz for artifacts, etc."""

import logging
import os
import shutil
import tarfile
import tempfile
from collections.abc import Iterator
from pathlib import Path

from debusine.db.models import Artifact, File


class TarArtifact:
    """Create a .tar.gz file."""

    def __init__(
        self, artifact: Artifact, subdirectory: str | None = None
    ) -> None:
        """Initialize member variables."""
        self._artifact = artifact

        self._files = self._get_file_list(artifact, subdirectory)
        self._pos = 0
        self._max_chunk_size = 50 * 1024 * 1024

        self._temp_directory = Path(
            tempfile.mkdtemp(prefix=f"debusine-artifact-download-{artifact.id}")
        )

        tar_file = self._temp_directory / f"artifact-{artifact.id}.tar.gz"

        self._tar = tarfile.open(tar_file.as_posix(), mode="w|gz")  # noqa: SIM115
        self._tar_closed = False

        self._tar_file = tar_file.open("rb")

        self._block_size = os.fstat(self._tar_file.fileno()).st_blksize
        self._bytes_deleted = 0

        self._closed = False

    @staticmethod
    def _get_file_list(
        artifact: Artifact, subdirectory: str | None
    ) -> dict[str, File]:
        """Return dictionary of file paths - file objects to be included."""
        files = {}

        if subdirectory is None:
            subdirectory = ""

        for file_in_artifact in (
            artifact.fileinartifact_set.select_related("file")
            .filter(path__startswith=subdirectory, complete=True)
            .order_by("-path")
        ):
            files[file_in_artifact.path] = file_in_artifact.file

        return files

    def __iter__(self) -> Iterator[bytes]:
        """Self is an iterator."""
        return self

    def __next__(self) -> bytes:
        """
        Return data or raise StopIteration if any exception happened.

        StopIteration if any exception occurred is needed to make debusine
        server to close the connection with the client and to avoid the
        client waiting for new data when no new data will be generated.

        The exception is logged for debugging purposes.
        """
        try:
            return self.next()
        except Exception as exc:
            if not isinstance(exc, StopIteration):
                logging.exception(exc)  # noqa: G200
            self.close()
            raise StopIteration

    def flush_tar(self) -> None:
        """Close the tar file so that the last pending data gets written."""
        self._tar.close()
        self._tar_closed = True

    def next(self) -> bytes:
        """Return pending data (from the last added file) and adds a file."""
        data = self._tar_file.read(self._max_chunk_size)

        if data != b"":
            # No need to add files, close, etc. if data can be
            # returned
            return data

        # No data read: add files, close Tar, StopIterating, etc.
        if len(self._files) > 0:
            # No data read and there are new files to add
            path, fileobj = self._files.popitem()

            file_backend = self._artifact.workspace.scope.download_file_backend(
                fileobj
            )
            with file_backend.get_stream(fileobj) as file:
                tarinfo = tarfile.TarInfo(path)
                tarinfo.size = fileobj.size
                mtime = self._artifact.created_at.timestamp()
                tarinfo.mtime = mtime
                self._tar.addfile(tarinfo, file)
        elif self._tar_closed:
            # No data read and the tar file is closed: stop iterating
            raise StopIteration
        elif len(self._files) == 0:
            # Once the tar file is finalized, some more data might be read on
            # the next iteration
            self.flush_tar()
        else:
            raise AssertionError  # pragma: no cover

        return b""

    def close(self) -> None:
        """Release all resources."""
        if self._closed:
            return
        if not self._tar_closed:
            self.flush_tar()
        self._tar_file.close()
        shutil.rmtree(self._temp_directory)
        self._closed = True



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
A fake FTP server for testing.

Do not use this in production!  It has only the features required for the
tests that currently use it, and no security.
"""

from pathlib import Path
from threading import Event, Thread
from typing import override

from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler
from pyftpdlib.ioloop import IOLoop
from pyftpdlib.servers import ThreadedFTPServer


# mypy complains that IOLoop is untyped, which is true, but we can't fix
# that here.
class InterruptibleIOLoop(IOLoop):  # type: ignore[misc]
    """A poller that can be interrupted by an event."""

    stop_event: Event | None = None

    def poll(self, timeout: float) -> None:
        """Poll once."""
        if (
            self.stop_event is not None and self.stop_event.is_set()
        ):  # pragma: no cover
            return
        super().poll(timeout)


class FakeFTPServerThread(Thread):
    """Thread that runs a fake FTP server for testing."""

    def __init__(self, username: str, home_directory: Path) -> None:
        """Initialize the server and start listening."""
        super().__init__()
        self.name = "FakeFTPServer"
        authorizer = DummyAuthorizer()
        authorizer.add_user(username, "", str(home_directory), perm="ew")
        handler = FTPHandler
        handler.authorizer = authorizer
        ioloop_class = InterruptibleIOLoop
        ioloop_class.stop_event = self.stop_event = Event()
        self.server = ThreadedFTPServer(("", 0), handler, ioloop=ioloop_class())

    @property
    def address(self) -> tuple[str, int]:
        """Return the (host, port) to be used by clients."""
        host, port = self.server.socket.getsockname()
        assert isinstance(host, str)
        assert isinstance(port, int)
        return host, port

    @override
    def run(self) -> None:
        """Run the server."""
        self.server.serve_forever()

    def stop(self) -> None:
        """Stop the server."""
        self.stop_event.set()
        self.server.close_all()
        self.join()



# Generated by Django 5.2.7 on 2025-12-24 16:39

from django.db import migrations, models

import debusine.artifacts.models
import debusine.tasks.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0037_collection_fields_help_text'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='worker',
            name='db_worker_celery_or_token',
        ),
        migrations.AddConstraint(
            model_name='worker',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('worker_type', debusine.tasks.models.WorkerType['CELERY']),
                    ('activation_token__isnull', False),
                    ('token__isnull', False),
                    models.Q(
                        ('instance_created_at__isnull', True),
                        ('worker_pool_id__isnull', False),
                    ),
                    _connector='OR',
                ),
                name='db_worker_celery_or_token',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving server tasks."""

import abc
from typing import Any, override

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.db.models import WorkRequest
from debusine.db.tasks import DBTask
from debusine.db.tasks.mock_resolver import MockFieldResolver
from debusine.server.tasks.base import BaseServerTask
from debusine.server.tasks.wait import BaseWaitTask
from debusine.tasks import inputs
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData
from debusine.tasks.tests.helpers import CurrentTaskAccessor, TaskTestCase
from debusine.test.django import TestCase


class DBTaskTestCase[TaskClass: DBTask[Any, Any]](
    TaskTestCase[TaskClass], TestCase, abc.ABC
):
    """Common base for DBTask tests."""

    #: Field resolver for the test
    resolver: MockFieldResolver

    #: Shortcut to self.task.work_request
    work_request = CurrentTaskAccessor[WorkRequest]()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver = MockFieldResolver()

    @abc.abstractmethod
    def work_request_create(self, **kwargs: Any) -> WorkRequest:
        """Set up the WorkRequest for the current task."""

    @override
    def task_create(self, **kwargs: Any) -> TaskClass:
        work_request = self.work_request_create(**kwargs)
        task = work_request.get_task()
        assert isinstance(task, self.task_class)
        return task

    @override
    def task_set_current(self, task: TaskClass) -> None:
        super().task_set_current(task)
        self.__dict__["work_request"] = task.work_request

    @override
    def task_teardown(self, task: TaskClass) -> None:
        if task._debug_log_files_directory:
            task._debug_log_files_directory.cleanup()
        super().task_teardown(task)

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        """Invoke compute_scheduler_tags."""
        self.task.resolve_inputs(inputs.Stage.PENDING)
        return self.task.compute_scheduler_tags()

    # TODO: this currently cannot use self.task because it cannot be typed to
    # return the task's dynamic data type. Passing self.task is shorter to type
    # than casting or passing the dynamic data type
    def compute_dynamic_data[DTD: BaseDynamicTaskData](
        self,
        task: DBTask[Any, DTD] | None = None,
        strip_task_inputs: bool = False,
    ) -> DTD:
        """Compute dynamic data for this task."""
        task = task or self.task
        task.resolve_inputs(inputs.Stage.RUNNING)
        ddata = task.compute_dynamic_data()
        if strip_task_inputs:
            ddata.task_input_fields = {}
        return ddata

    @override
    def build_expected_tag_candidates(
        self, **kwargs: set[str]
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        workspace = self.task.workspace
        provided, required = super().build_expected_tag_candidates(**kwargs)
        provided.add(
            ttags.ProvenanceProvided.SYSTEM,
            {
                ttags.SCOPE_PREFIX + workspace.scope.name,
                ttags.WORKSPACE_PREFIX
                + f"{workspace.scope.name}:{workspace.name}",
            },
        )
        return provided, required


class ServerTaskTestCase[TaskClass: BaseServerTask[Any, Any]](
    DBTaskTestCase[TaskClass], TestCase
):
    """Common base for server task tests."""

    #: Set to True for the task to be marked running
    mark_running: bool

    @override
    def setUp(self) -> None:
        super().setUp()
        self.mark_running = True

    @override
    def work_request_create(self, **kwargs: Any) -> WorkRequest:
        task_data = self.task_data.copy()
        task_data.update(kwargs)
        work_request = self.playground.create_server_task(
            task_name=self.task_class.name,
            task_data=self.task_class.task_data_type(**task_data),
        )
        self.playground.advance_work_request(
            work_request, mark_running=self.mark_running
        )
        return work_request

    @override
    def build_expected_tag_candidates(
        self, **kwargs: set[str]
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().build_expected_tag_candidates(**kwargs)
        required.add(
            wtags.ProvenanceRequired.SYSTEM,
            {
                wtags.WORKER_TYPE_CELERY,
                wtags.TASK_PREFIX
                + f"server:{self.task.name}:version:{self.task.TASK_VERSION}",
            },
        )
        return provided, required


class SampleBaseDBTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    DBTask[TD, DTD]
):
    """Common test implementation of abstract task methods."""

    @override
    def compute_dynamic_data(self) -> DTD:
        """Resolve artifact lookups for this task."""
        return self.dynamic_task_data_type()


class SampleBaseServerTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    SampleBaseDBTask[TD, DTD], BaseServerTask[TD, DTD]
):
    """Common test implementation of BaseServerTask methods."""


class SampleBaseWaitTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    SampleBaseDBTask[TD, DTD], BaseWaitTask[TD, DTD]
):
    """Common test implementation of BaseWaitTask methods."""
